"""
This submodule contains several pre-processing techniques for gene expression datasets (standardizations, discretizations and RNAseq normalization)
"""
__author__ = "Sergio Peignier and Pauline Schmitt"
__copyright__ = "Copyright 2019, The GReNaDIne Project"
__license__ = "GPL"
__version__ = "0.0.9"
__maintainer__ = "Sergio Peignier"
__email__ = "sergio.peignier@insa-lyon.fr"
__status__ = "pre-alpha"



# -*- coding: utf-8 -*-
"""
This module allows to visualize, plot and compute metrics on the infered gene regulatory networks, using built-in metrics in `networkx`_.

.. _scikit-learn:
    https://scikit-learn.org
"""
import pandas as pd
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from scipy import stats
from grenadine.Inference.inference import rank_GRN
from sklearn.decomposition import PCA
from sklearn.metrics import roc_curve
from sklearn.metrics import roc_auc_score
from sklearn.metrics import auc
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from sklearn.metrics import accuracy_score
from sklearn.metrics import f1_score
from tqdm import tqdm
from seaborn import color_palette,clustermap,boxplot
from matplotlib.patches import Patch


def plot_scores(scores_, metrics_list = None):
    """
    Plots the evaluation scores distribution as boxplots

    Args:
        metrics_list (list): list of metrics names in scores_ to be represented
    """
    if metrics_list is None:
        metrics_list = [c for c in scores_.columns if "mean_test" in c]
    s = scores_[metrics_list]
    boxplot(s,
            orient='h',notch=True, showcaps=False,
            flierprops={"marker": "."},
            boxprops={"facecolor": (.1, .1, .1, .1)},
            medianprops={"color": "darkred", "linewidth": 3})
    plt.xlabel('Mean best model score',fontsize=10)
    plt.xlabel('')
    _=plt.yticks(fontsize=12)
    _=plt.xticks(fontsize=12)
    plt.xlim([-0.03,1.03])


def clustermap_with_classes(df,classes_row, classes_col,**clustermap_params):
    row_colors = None
    col_colors = None
    if classes_row is not None:
        unique_classes_row =  classes_row.unique()
        unique_classes_row.sort()
        colors_row = color_palette("Greens",n_colors=len(unique_classes_row))
        colors_map_row = {c:colors_row[i] for i,c in enumerate(unique_classes_row)} 
        row_colors = classes_row.map(colors_map_row)

    if classes_col is not None:
        unique_classes_col =  classes_col.unique()
        unique_classes_col.sort()
        colors_col = color_palette("Blues",n_colors=len(unique_classes_col))
        colors_map_col = {c:colors_col[i] for i,c in enumerate(unique_classes_col)} 
        col_colors = classes_col.map(colors_map_col)
    
    default_clustermap_params = {'center':0,
                                 'method':'ward',
                                 'xticklabels':False,
                                 'yticklabels':False,
                                 'figsize':(3,3),
                                }
    
    default_clustermap_params.update(clustermap_params)
    
    ax = clustermap(df,
                    row_colors=row_colors,
                    col_colors=col_colors,
                    **default_clustermap_params,
                    )
    if classes_col is not None:
        handles = [Patch(facecolor=colors_map_col[name]) for name in colors_map_col]
        legend_col =plt.legend(handles, 
                               colors_map_col, 
                               title="Column class",
                               bbox_to_anchor=(-0., 1),
                               bbox_transform=plt.gcf().transFigure,
                               loc='upper right')
        plt.gca().add_artist(legend_col)

    if classes_row is not None:
        handles = [Patch(facecolor=colors_map_row[name]) for name in colors_map_row]
        legend_row =plt.legend(handles, 
                               colors_map_row, 
                               title="Row class",
                               bbox_to_anchor=(-0., 0.6),
                               bbox_transform=plt.gcf().transFigure,
                               loc='upper right')
        plt.gca().add_artist(legend_row)
    return ax


def to_graph(ranking_df):
    """
    This method returns a directed graph generated by networkx
    """
    return(nx.from_pandas_edgelist(ranking_df,source='TF',target='TG',edge_attr='score'))

def draw_by(G, dico, maxi):
    colnodes = list()
    sizenodes = list()
    for node in G.nodes:
        colnodes.append(dico[node])
        sizenodes.append((120.*dico[node])/maxi)
    nx.draw(G, pos = nx.nx_agraph.graphviz_layout(G), node_color = colnodes, node_size = sizenodes, cmap = 'plasma')


def rich_club_coefficient(G,show=False):
    undir_G=nx.to_undirected(G)
    rich_club = nx.rich_club_coefficient(undir_G)
    if(show):
        plt.bar(*zip(*rich_club.items()))
        plt.xlabel('k minimum')
        plt.ylabel('Coefficient rich club')
        plt.show()
    return(rich_club)


def betweenness_centrality(G,show=False):
    between = nx.betweenness_centrality(G)
    if(show):
        plt.bar(*zip(*between.items()))
        plt.xlabel('gene')
        plt.ylabel('Betweeness centrality')
        plt.show()
    return(between)

def clustering_coefficient(G,show=False):
    undir_G=nx.to_undirected(G)
    clust = nx.clustering(undir_G)
    if(show):
        plt.bar(*zip(*clust.items()))
        plt.xlabel('gene')
        plt.ylabel('Clustering coefficient')
        plt.show()
    return(clust)

def closeness_centrality(G,show=False):
    undir_G=nx.to_undirected(G)
    closeness = nx.closeness_centrality(undir_G)
    if(show):
        plt.bar(*zip(*closeness.items()))
        plt.xlabel('gene')
        plt.ylabel('Closeness centrality')
        plt.show()
    return(closeness)



"""
This module allows to infer Gene Regulatory Networks using
gene expresion data (RNAseq or Microarray). This module implements several
inference algorithms based on classification, using `scikit-learn`_.

.. _scikit-learn:
    https://scikit-learn.org
"""

import pandas as pd
import numpy as np
from sklearn.ensemble import RandomForestClassifier as _sklearn_RandomForestClassifier
from sklearn.ensemble import ExtraTreesClassifier as _sklearn_ExtraTreesClassifier
from sklearn.ensemble import AdaBoostClassifier as _sklearn_AdaBoostClassifier
from sklearn.ensemble import GradientBoostingClassifier as _sklearn_GradientBoostingClassifier
from sklearn.svm import LinearSVC as _sklearn_SVC
from sklearn.naive_bayes import MultinomialNB as _sklearn_MultinomialNB
from sklearn.naive_bayes import ComplementNB as _sklearn_ComplementNB
from sklearn.ensemble import BaggingClassifier as _sklearn_BaggingClassifier
from collections import Counter

def RF_classifier_score(X, y, **rf_parameters):
    """
    Random Forest Classifier, score predictor function based on `scikit-learn`_
    RandomForestClassifier.

    Args:
        X (pandas.DataFrame): Transcription factor gene expressions (discretized
            or not) where rows are experimental conditions and columns are
            transcription factors
        y (pandas.Series): Target gene expression vector (discretized) where
            rows are experimental conditions
        **rf_parameters: Named parameters for the sklearn _sklearn_RandomForestClassifier

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        RandomForestClassifier to the regulatory relationship between the target
        gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randint(0,3,size=5), index=["c1","c2","c3","c4","c5"])
        >>> scores = RF_classifier_score(tfs,tg)
        >>> scores
        array([0.21071429, 0.4       , 0.28928571])
    """
    classifier = _sklearn_RandomForestClassifier(**rf_parameters)
    classifier.fit(X, y)
    scores = classifier.feature_importances_
    return scores

def XRF_classifier_score(X, y, **xrf_parameters):
    """
    Randomized decision trees Classifier, score predictor function based on
    `scikit-learn`_ ExtraTreesClassifier.

    Args:
        X (pandas.DataFrame): Transcription factor gene expressions (discretized
            or not) where rows are experimental conditions and columns are
            transcription factors
        y (pandas.Series): Target gene expression vector (discretized) where
            rows are experimental conditions
        **xrf_parameters: Named parameters for the sklearn _sklearn_ExtraTreesClassifier

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        ExtraTreesClassifier to the regulatory relationship between the target
        gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randint(0,3,size=5), index=["c1","c2","c3","c4","c5"])
        >>> scores = XRF_classifier_score(tfs,tg)
        >>> scores
        array([0.31354167, 0.35520833, 0.33125   ])
    """
    classifier = _sklearn_ExtraTreesClassifier(**xrf_parameters)
    classifier.fit(X, y)
    scores = classifier.feature_importances_
    return(scores)

def AdaBoost_classifier_score(X, y, **adab_parameters):
    """
    AdaBoost Classifier, score predictor function based on `scikit-learn`_
    AdaBoostClassifier.

    Args:
        X (pandas.DataFrame): Transcription factor gene expressions (discretized
            or not) where rows are experimental conditions and columns are
            transcription factors
        y (pandas.Series): Target gene expression vector (discretized) where
            rows are experimental conditions
        **adab_parameters: Named parameters for the sklearn AdaBoostClassifier

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        AdaBoostClassifier to the regulatory relationship between the target
        gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randint(0,3,size=5), index=["c1","c2","c3","c4","c5"])
        >>> scores = AdaBoost_classifier_score(tfs,tg)
        >>> scores
        array([0.24, 0.44, 0.32])
    """
    classifier = _sklearn_AdaBoostClassifier(**adab_parameters)
    classifier.fit(X, y)
    scores = classifier.feature_importances_
    return scores

def GB_classifier_score(X, y, **gb_parameters):
    """
    Gradient Boosting Classifier, score predictor function based on
    `scikit-learn`_ GradientBoostingClassifier.

    Args:
        X (pandas.DataFrame): Transcription factor gene expressions (discretized
            or not) where rows are experimental conditions and columns are
            transcription factors
        y (pandas.Series): Target gene expression vector (discretized) where
            rows are experimental conditions
        **gb_parameters: Named parameters for the sklearn _sklearn_ExtraTreesClassifier

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        GradientBoostingClassifier to the regulatory relationship between the target
        gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randint(0,3,size=5), index=["c1","c2","c3","c4","c5"])
        >>> scores = GB_classifier_score(tfs,tg)
        >>> scores
         array([0.33959125, 0.21147015, 0.4489386 ])
    """
    classifier = _sklearn_GradientBoostingClassifier(**gb_parameters)
    classifier.fit(X, y)
    scores = classifier.feature_importances_
    return scores

def SVM_classifier_score(X, y, **svm_parameters):
    """
    SVM Classifier, score predictor function based on `scikit-learn`_ SVC
    (Support Vector Classifier).

    Args:
        X (pandas.DataFrame): Transcription factor gene expressions (discretized
            or not) where rows are experimental conditions and columns are
            transcription factors
        y (pandas.Series): Target gene expression vector (discretized) where
            rows are experimental conditions
        **svm_parameters: Named parameters for the sklearn SVC

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        SVC to the regulatory relationship between the target
        gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randint(0,3,size=5), index=["c1","c2","c3","c4","c5"])
        >>> scores = SVM_classifier_score(tfs,tg)
        >>> scores
        array([0.58413783, 0.5448345 , 0.31764191])
    """
    # consider changing kernel function: SVC(kernel = 'linear'/'poly'/'sigmoid'/'precomputed')
    # default is rbf
    # also consider changing polynomial degree : SVC(kernel = 'poly', degree = int value)
    # default is 3
    # but linear is the only kernel that has coef_ attribute
    #svm_parameters["kernel"] = 'linear'
    svm_parameters["multi_class"] = 'ovr'
    svm_parameters["dual"]=False
    classifier = _sklearn_SVC(**svm_parameters)
    classifier.fit(X, y)
    # coef_ = array of shape (nb_classes, nb_TFs)
    # replacing each class value (= discretized value) of y with corresponding TFs weights from attribute coef_
    # then taking the mean along each column (= mean importance of TF over all conditions)
    if classifier.coef_.shape[0] > 1:
        scores = np.abs(classifier.coef_).mean(axis=0)
        #coef = pd.DataFrame(classifier.coef_,index=classifier.classes_)
        #nb_classes = pd.Series(Counter(y))
        #scores = np.abs((coef.T*nb_classes/nb_classes.sum()).T).mean(axis=0)
    else:
        scores = np.abs(classifier.coef_[0,:])
    return scores


def MultinomialNB_classifier_score(X, y, **nb_parameters):
    """
    Multinomial Naive Bayes Classifier, score predictor function based on
    `scikit-learn`_ MultinomialNB.

    Args:
        X (pandas.DataFrame): Transcription factor gene expressions (discretized
            or not) where rows are experimental conditions and columns are
            transcription factors
        y (pandas.Series): Target gene expression vector (discretized) where
            rows are experimental conditions
        **nb_parameters: Named parameters for the sklearn MultinomialNB

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        MultinomialNB to the regulatory relationship between the target
        gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randint(0,3,size=5), index=["c1","c2","c3","c4","c5"])
        >>> scores = MultinomialNB_classifier_score(tfs,tg)
        >>> scores
        array([0.3010284 , 0.41871716, 0.4272386 ])
    """
    x_shift = 0
    x_min = X.min().min()
    if x_min<0:
        x_shift = -x_min
    classifier = _sklearn_MultinomialNB(**nb_parameters)
    classifier.fit(X+x_shift, y)
    # coef_ = array of shape (nb_classes, nb_TFs)
    # replacing each class value (= discretized value) of y with corresponding TFs weights from attribute coef_
    # then taking the mean along each column (= mean importance of TF over all conditions)
    if classifier.coef_.shape[0] > 1:
        coef = pd.DataFrame(classifier.feature_log_prob_,index=classifier.classes_)
        scores = np.abs(coef)
        scores = coef.fillna(0)
        scores = scores.mean(axis=0)
    else:
        scores = np.abs(classifier.feature_log_prob_[0,:])
        scores = np.nan_to_num(scores)
        #scores = scores.fillna(0)
    return np.array(scores)


def ComplementNB_classifier_score(X, y, **nb_parameters):
    """
    Complement Naive Bayes Classifier, score predictor function based on
    `scikit-learn`_ ComplementtNB.

    Args:
        X (pandas.DataFrame): Transcription factor gene expressions (discretized
            or not) where rows are experimental conditions and columns are
            transcription factors
        y (pandas.Series): Target gene expression vector (discretized) where
            rows are experimental conditions
        **nb_parameters: Named parameters for the sklearn MultinomialNB

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        ComplementNB to the regulatory relationship between the target
        gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randint(0,3,size=5), index=["c1","c2","c3","c4","c5"])
        >>> scores = ComplementNB_classifier_score(tfs,tg)
        >>> scores
        array([0.28113447, 0.39096368, 0.45629413])
    """
    x_shift = 0
    x_min = X.min().min()
    if x_min<0:
        x_shift = -x_min
    classifier = _sklearn_ComplementNB(**nb_parameters)
    classifier.fit(X+x_shift, y)
    # coef_ = array of shape (nb_classes, nb_TFs)
    # replacing each class value (= discretized value) of y with corresponding
    # TFs weights from attribute coef_
    # then taking the mean along each column (= mean importance of TF over all
    # conditions)
    if classifier.coef_.shape[0] > 1:
        coef = pd.DataFrame(classifier.feature_log_prob_,index=classifier.classes_)
        scores = np.abs(coef)
        scores = coef.fillna(0)
        scores = scores.mean(axis=0)
    else:
        scores = np.abs(classifier.feature_log_prob_[0,:])
        scores = np.nan_to_num(scores)
        #scores = scores.fillna(0)
    return np.array(scores)


def bagging_classifier_score(X, y, **bagging_parameters):
    """
    Apply the bagging technique to a regression algorithm, based on
    `scikit-learn`_ BaggingClassifier.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **adab_parameters: Named parameters for the sklearn AdaBoostRegressor

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the average score
        assigned by the Base Regressor to the regulatory relationship
        between the target gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> from sklearn.svm import SVR
        >>> np.random.seed(0)
        >>> svc = SVC(kernel="linear",decision_function_shape='ovr')
        >>> nb_conditions = 10
        >>> tfs = pd.DataFrame(np.random.randn(nb_conditions,3),
                       index =["c"+str(i) for i in range(nb_conditions)],
                       columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randint(0,2,size=nb_conditions),
                           index =["c"+str(i) for i in range(nb_conditions)])
        >>> bagging_parameters = {"base_estimator":svc,
                                  "n_estimators":5,
                                  "max_samples":0.9}
        >>> scores = bagging_classifier_score(tfs,tg,**bagging_parameters)
        >>> scores
        array([0.269231,0.412219,0.299806])
    """
    def get_score(classifier):
        scores=None
        if hasattr(classifier, 'feature_importances_'):
            scores = pd.DataFrame(feature_importances).mean(axis=0)
        elif hasattr(classifier,"coef_"):
            if classifier.coef_.shape[0] > 1:
                scores = np.abs(classifier.coef_).mean(axis=0)
            else:
                scores = np.abs(classifier.coef_[0,:])
        return(scores)
    bc = _sklearn_BaggingClassifier(**bagging_parameters)#BaggingClassifier
    bc.fit(X,y)
    coefs = [get_score(e) for e in bc.estimators_]
    coefs = [c for c in coefs if c is not None]
    if len(coefs):
        scores = pd.DataFrame(coefs).mean(axis=0)
    else:
        print("Base regressor has not a feature_importance of coef_ attribute")
        return(0)
    return(scores)



"""
This module allows to infer co-expression Gene Regulatory Networks using
gene expression data (RNAseq or Microarray). This module implements severall
inference algorithms based on statistical predictors, using `scipy-stats`_ and
`scikit-learn`_.

.. _scipy-stats:
    https://docs.scipy.org/doc/scipy/reference/stats.html
.. _scikit-learn:
    https://scikit-learn.org
"""

from sklearn.feature_selection import f_regression as _sklearn_f_regression
from sklearn.feature_selection import mutual_info_regression as _sklearn_mutual_info_regression
from scipy.stats import spearmanr as _scipy_spearmanr
from scipy.stats import pearsonr as _scipy_pearsonr
from scipy.stats import wilcoxon as _scipy_wilcoxon
from scipy.stats import mannwhitneyu as _scipy_mannwhitneyu
from scipy.stats import theilslopes as _scipy_theilslopes
from scipy.stats import kendalltau as _scipy_kendalltau
from scipy.stats import rankdata as _scipy_rankdata
from scipy.stats import energy_distance as _scipy_energy_distance
from scipy.stats import wasserstein_distance as _scipy_wasserstein_distance
import numpy as np


def abs_pearsonr_coef(X,y):
    """
    Score predictor function based on the `scipy-stats`_
    absolute Pearson correlation.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the absolute value of the
        correlation between target gene expression and the i-th transcription
        factor gene expression.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = abs_pearsonr_coef(tfs,tg)
        >>> scores
        array([0.41724166, 0.02212467, 0.23708491])
    """
    scores = np.zeros(X.shape[1])
    for i in range(X.shape[1]):
        x_tf = X[:,i]
        scores[i] = _scipy_pearsonr(y,x_tf)[0]
    scores = np.abs(scores)
    return(scores)

def abs_spearmanr_coef(X,y):
    """
    Score predictor function based on the `scipy-stats`_
    absolute Spearman correlation.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the absolute value of the
        correlation between target gene expression and the i-th transcription
        factor gene expression.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = abs_spearmanr_coef(tfs,tg)
        >>> scores
        array([0.5, 0.3, 0.3])
    """
    scores = np.zeros(X.shape[1])
    for i in range(X.shape[1]):
        x_tf = X[:,i]
        scores[i] = _scipy_spearmanr(y,x_tf)[0]
    scores = np.abs(scores)
    return(scores)

def kendalltau_score(X,y,**kendalltau_parameters):
    """
    Score predictor function based on the `scipy-stats`_
    Kendall’s tau correlation measure.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **kendalltau_parameters: Named parameters for the scipy-stats kendall's
            tau correlation measure

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score of the
        score between target gene expression and the i-th transcription factor
        gene expression.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = kendalltau_score(tfs,tg)
        >>> scores
        array([0.8487997 , 1.30065214, 0.20467198])s
    """
    epsilon = 1e-300
    scores = np.zeros(X.shape[1])
    y = _scipy_rankdata(y)
    for i in range(X.shape[1]):
        x_tf = X[:,i]
        x_tf_ranks = _scipy_rankdata(x_tf)
        scores[i] = _scipy_kendalltau(y,x_tf_ranks,**kendalltau_parameters)[1]
        scores[i] = -np.log10(scores[i]+epsilon)
    return(scores)

def f_regression_score(X,y):
    """
    Score predictor function based on the `scikit-learn`_ f_regression score.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score of the
        f_regression linear test between target gene expression and the
        i-th transcription factor gene expression.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = f_regression_score(tfs,tg)
        >>> scores
        array([0.63235967, 0.00146922, 0.17867071])
    """
    scores, p_values = _sklearn_f_regression(X, y, center=True)
    return(scores)

def CLR(X,y,**mi_parameters):
    """
    Score predictor function based on `scikit-learn`_ mutual_info_regression
    score.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **mi_parameters: Named parameters for sklearn mutual_info_regression

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score of the
        sklearn mutual_info_regression computation between target gene
        expression and the i-th transcription factor gene expression.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = CLR(tfs,tg)
        >>> scores
        array([6.66666667e-02, 1.16666667e-01, 2.22044605e-16])
    """
    scores = _sklearn_mutual_info_regression(X,y,**mi_parameters)
    return(scores)

def wilcoxon_score(X,y,**wilcoxon_parameters):
    """
    Score predictor function based on the `scipy-stats`_
    Wilcoxon signed-rank test.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **wilcoxon_parameters: Named parameters for the scipy-stats Wilcoxon
            signed-rank test

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score
        between target gene expression and the i-th transcription
        factor gene expression.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),index =["c1","c2","c3","c4","c5"],columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = wilcoxon_score(tfs,tg)
        >>> scores
        array([1.36537718, 0.64797987, 0.30086998])
    """
    epsilon = 1e-300
    scores = np.zeros(X.shape[1])
    for i in range(X.shape[1]):
        x_tf = X[:,i]
        scores[i] = _scipy_wilcoxon(y,x_tf,**wilcoxon_parameters)[1]
        scores[i] = -np.log10(scores[i]+epsilon)
    return(scores)

def mannwhitneyu_score(X,y,**mannwhitneyu_parameters):
    """
    Score predictor function based on the `scipy-stats`_ Mann-Whitney rank test.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **mannwhitneyu_parameters: Named parameters for the scipy-stats
            Mann-Whitney rank test

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score
        between target gene expression and the i-th transcription
        factor gene expression.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = mannwhitneyu_score(tfs,tg)
        >>> scores
        array([1.52213525, 0.47101693, 0.3795872 ])
    """
    epsilon = 1e-300
    scores = np.zeros(X.shape[1])
    for i in range(X.shape[1]):
        x_tf = X[:,i]
        scores[i] = _scipy_mannwhitneyu(y,x_tf,**mannwhitneyu_parameters)[1]
        scores[i] = -np.log10(scores[i]+epsilon)
    return(scores)

def theilslopes_score(X,y,**theilslopes_parameters):
    """
    Score predictor function based on the `scipy-stats`_
    Theil-Sen robust slope estimator.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **theilslopes_parameters: Named parameters for the scipy-stats
            Theil-Sen robust slope estimator

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score
        between target gene expression and the i-th transcription
        factor gene expression.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = theilslopes_score(tfs,tg)
        >>> scores
        array([0.92309299, 0.90933202, 0.26451817])
    """
    scores = np.zeros(X.shape[1])
    for i in range(X.shape[1]):
        x_tf = X[:,i]
        scores[i] = np.abs(_scipy_theilslopes(y,x_tf,**theilslopes_parameters)[0])
    return(scores)


def energy_distance_score(X,y,**energy_distance_parameters):
    """
    Score predictor function based on the `scipy-stats`_ energy distance between
    1D distributions.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **energy_distance_parameters: Named parameters for the scipy-stats
            energy distance

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score
        between target gene expression and the i-th transcription
        factor gene expression.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = energy_distance_score(tfs,tg)
        >>> scores
        array([0.40613705, 0.6881455 , 0.72786711])
    """
    scores = np.zeros(X.shape[1])
    for i in range(X.shape[1]):
        x_tf = X[:,i]
        scores[i] = np.exp(-_scipy_energy_distance(y,x_tf,**energy_distance_parameters))
    return(scores)

def wasserstein_distance_score(X,y,**wasserstein_distance_parameters):
    """
    Score predictor function based on the `scipy-stats`_ Wasserstein distance
    between 1D distributions.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **wasserstein_distance_parameters: Named parameters for the scipy-stats
            Wasserstein distance

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score
        between target gene expression and the i-th transcription
        factor gene expression.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = wasserstein_distance_score(tfs,tg)
        >>> scores
        array([0.36457586, 0.72057084, 0.81207932])
    """
    scores = np.zeros(X.shape[1])
    for i in range(X.shape[1]):
        x_tf = X[:,i]
        scores[i] = np.exp(-_scipy_wasserstein_distance(y,x_tf,**wasserstein_distance_parameters))
    return(scores)



# -*- coding: utf-8 -*-
"""
This module allows to build predictive GRNs models using regression methods
"""
from tqdm.autonotebook import tqdm
import pandas as pd
import numpy as np
import networkx as nx
from sklearn.model_selection import cross_val_score
from sklearn.linear_model import BayesianRidge
from grenadine.Evaluation.evaluation import get_top_tfs_per_tg_from_scores
import copy
import random
from deap import base
from deap import creator
from deap import tools
creator.create("FitnessMax", base.Fitness, weights=(1.0,))
creator.create("Individual", list, fitness=creator.FitnessMax)


def ranking_to_networkx(ranking_df):
    """
    This method returns a directed graph generated by networkx

    Args:
        ranking_df (pandas.DataFrame): ranking matrix.
        A ranking matrix contains a row for each possible regulatory link, it
        also contains 4 columns, namely the rank, the score, the transcription
        factor id, and the target gene id.

    Returns:
        networkx.Digraph: networkx gene regulatory network

    """
    return(nx.from_pandas_edgelist(ranking_df,source='TF',target='TG'))

def join_models(*models):
    model = models[0]
    for i in range(1,len(models)):
        model.tgs_tfs.update(models[i].tgs_tfs)
        model.M.update(models[i].M)
        model.eval.update(models[i].eval)
    return model


class Model:
    def __init__(self,
                 tgs_tfs,
                 predictor_class,
                 **predictor_parameters):
        self.tgs_tfs = tgs_tfs
        self.clf = predictor_class(**predictor_parameters)
        self.M = {tg:None for tg in self.tgs_tfs}

    def flatten_evaluation(self):
        evaluation = []
        for c in self.eval:
            for tg in self.eval[c]:
                for s in self.eval[c][tg]:
                    evaluation.append([c,tg,s])
        evaluation = pd.DataFrame(evaluation,columns=["condition","TG","score"])
        return(evaluation)

    def evaluation_tg(self,tg):
        eval_tg = pd.DataFrame({c: m2.eval[c][tg] for c in m2.eval}).T
        eval_tg_mean = eval_tg.mean(axis=1)
        return eval_tg.loc(eval_tg_mean.sort_values(ascending=False).index)

    def evaluation_avg(self):
        eval_avg = pd.DataFrame({c: {tg: m2.eval[c][tg].mean() for tg in m2.eval[c] } for c in m2.eval})
        return(eval_avg)

    def _get_coefs_tg(self,tg):
        coefs = self._get_coefs()
        return pd.DataFrame({c: {tf: coefs[c][tg][tf] for tf in coefs[c][tg] } for c in coefs})

    def fit(self, gene_expression_matrix, conditions=None,cv=3):
        if conditions is not None:
            self.M = {c:{tg:None for tg in self.tgs_tfs} for c in set(conditions)}
            self.eval = {c:{tg:None for tg in self.tgs_tfs} for c in set(conditions)}
            for c in tqdm(list(set(conditions))):
                gene_exp_local = gene_expression_matrix.loc[:,list(conditions==c)]
                for tg in self.tgs_tfs:
                    X = gene_exp_local.loc[self.tgs_tfs[tg]].T
                    y = gene_exp_local.loc[tg]
                    self.M[c][tg] = copy.deepcopy(self.clf.fit(X=X,y=y))
                    self.eval[c][tg] = cross_val_score(self.clf,X,y,cv=cv)
        else:
            self.M = {tg:None for tg in self.tgs_tfs}
            self.eval = {tg:None for tg in self.tgs_tfs}
            for tg in self.tgs_tfs:
                X = gene_expression_matrix.loc[self.tgs_tfs[tg]].T
                y = gene_expression_matrix.loc[tg]
                self.M[tg] = self.clf.fit(X=X,y=y)
                self.eval[tg] = cross_val_score(self.clf,X,y,cv=cv)

    def predict(self, gene_expression_matrix, conditions=None):
        if conditions is not None:
            results = {c:{tg:0 for tg in self.tgs_tfs} for c in set(conditions)}
            for c in list(set(conditions)):
                gene_exp_local = gene_expression_matrix.loc[:,list(conditions==c)]
                for tg in self.tgs_tfs:
                    X = gene_exp_local.loc[self.tgs_tfs[tg]].T
                    if c in self.M:
                        results[c][tg] = self.M[c][tg].predict(X=X)
                    else:
                        results[c][tg] = self.M[tg].predict(X=X)
        else:
            results = {tg:0 for tg in self.tgs_tfs}
            for tg in self.tgs_tfs:
                X = gene_expression_matrix.loc[self.tgs_tfs[tg]].T
                results[tg] = self.M[tg].predict(X=X)
        return results

    def score(self, gene_expression_matrix, conditions=None):
        if conditions is not None:
            results = {c:{tg:0 for tg in self.tgs_tfs} for c in set(conditions)}
            for c in list(set(conditions)):
                gene_exp_local = gene_expression_matrix.loc[:,list(conditions==c)]
                for tg in self.tgs_tfs:
                    X = gene_exp_local.loc[self.tgs_tfs[tg]].T
                    y = gene_exp_local.loc[tg]
                    if c in self.M:
                        results[c][tg] = self.M[c][tg].score(X=X,y=y)
                    else:
                        results[c][tg] = self.M[tg].score(X=X,y=y)
        else:
            results = {tg:0 for tg in self.tgs_tfs}
            for tg in self.tgs_tfs:
                X = gene_expression_matrix.loc[self.tgs_tfs[tg]].T
                y = gene_expression_matrix.loc[tg]
                results[tg] = self.M[tg].score(X=X,y=y)
        return results

    def _get_coefs(self):
        scores = {}
        for c in self.M:
            scores[c] = {}
            if type(self.M[c]) == dict:
                for tg in self.M[c]:
                    if "coef_" in self.M[c][tg].__dict__:
                        scores[c][tg] = {tf: self.M[c][tg].coef_[i] for i,tf in enumerate(self.tgs_tfs[tg])}
            else:
                scores[c] = {tf: self.M[c].coef_[i] for i,tf in enumerate(self.tgs_tfs[c])}
        return(scores)


def Evolve_optimal_model_tg(X,
                            tfs_full,
                            tg,
                            alpha=0.5,
                            genome_size = 30,
                            population_size=50,
                            p_init=0.1,
                            p_mut=0.1,
                            tour_size=10,
                            p_indiv_mut=0.1,
                            p_cross_over=0.5,
                            nb_generations=20,
                            model=BayesianRidge):
    def generate_bit(p):
        if random.random()<p:
            return 1
        return 0

    def map_individual(individual,tfs_full=tfs_full):
        tfs = [tfs_full[i] for i,g in enumerate(individual) if g]
        return(tfs)

    def evalOneMax(individual,alpha=alpha,tg=tg,tfs_full=tfs_full,X=X,model=model):
        if sum(individual)==0:
            return -10,
        tfs = map_individual(individual)
        tgs_tfs = {tg:tfs}
        model = Model(tgs_tfs,model)
        model.fit(X)
        p_active_genes = sum(individual)/len(individual)
        fitness = model.score(X)[tg] - alpha*p_active_genes
        return(fitness,)

    # generate model
    toolbox = base.Toolbox()
    toolbox.register("attr_bool", generate_bit, p_init)
    toolbox.register("individual", tools.initRepeat, creator.Individual, toolbox.attr_bool, genome_size)
    toolbox.register("population", tools.initRepeat, list, toolbox.individual)
    toolbox.register("evaluate", evalOneMax)
    toolbox.register("mate", tools.cxTwoPoint)
    toolbox.register("mutate", tools.mutFlipBit, indpb=p_mut)
    toolbox.register("select", tools.selTournament, tournsize=tour_size)
    pop = toolbox.population(population_size)

    # Compute fitnesses
    fitnesses = list(map(toolbox.evaluate, pop))
    for ind, fit in zip(pop, fitnesses):
        ind.fitness.values = fit
    fits = [ind.fitness.values[0] for ind in pop]

    # distributions hall-of-fame
    fitness_distrib = fits[:]
    best_fitnesses = [max(fits)]
    hall_of_fame = [pop[fits.index(best_fitnesses[0])]]
    generations = [0 for i in range(len(fits))]
    # Begin the evolution
    for g in tqdm(range(nb_generations)):
        offspring = toolbox.select(pop, len(pop))
        offspring = list(map(toolbox.clone, offspring))
        # cross over
        for child1, child2 in zip(offspring[::2], offspring[1::2]):
            if random.random() < p_cross_over:
                toolbox.mate(child1, child2)
                del child1.fitness.values
                del child2.fitness.values
        # mutation
        for mutant in offspring:
            if random.random() < p_indiv_mut:
                toolbox.mutate(mutant)
                del mutant.fitness.values
        # Evaluate new individuals
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit
        # Replace population
        pop = offspring
        # Gather all the fitnesses in one list and print the stats
        fits = [ind.fitness.values[0] for ind in pop]
        best_fitness = max(fits)
        best_fitnesses.append(best_fitness)
        hall_of_fame.append(pop[fits.index(best_fitness)])
        fitness_distrib+=fits
        generations+=[g+1 for i in range(len(fits))]
    fitness_distrib = pd.DataFrame({"fitness":fitness_distrib,
                                    "generation": generations})
    hall_of_fame = list(map(map_individual,hall_of_fame))
    best_model = Model({tg:hall_of_fame[-1]},model)
    best_model.fit(X)
    return fitness_distrib,list(hall_of_fame),best_fitnesses,best_model

def Evolve_optimal_model(X,
                         scores,
                         nb_full_tfs=30,
                         alpha=0.5,
                         genome_size = 30,
                         population_size=50,
                         p_init=0.1,
                         p_mut=0.1,
                         tour_size=10,
                         p_indiv_mut=0.1,
                         p_cross_over=0.5,
                         nb_generations=20,
                         model=BayesianRidge):
    top_tfs = get_top_tfs_per_tg_from_scores(scores,nb_full_tfs)
    models = []
    fitnesses_full = {}
    hall_of_fame_full = {}
    best_fitnesses_full = {}
    for i,tg in enumerate(top_tfs.index):
        print(tg+" "+str(i)+"/"+str(top_tfs.shape[0]))
        results = Evolve_optimal_model_tg(X,
                                          list(top_tfs.loc[tg]),
                                          tg,
                                          alpha,
                                          genome_size ,
                                          population_size,
                                          p_init,
                                          p_mut,
                                          tour_size,
                                          p_indiv_mut,
                                          p_cross_over,
                                          nb_generations,
                                          model)
        fitnesses,hall_of_fame,best_fitnesses,best_model = results
        fitnesses_full[tg] = fitnesses
        hall_of_fame_full[tg] = hall_of_fame
        best_fitnesses_full[tg] = best_fitnesses
        models.append(best_model)
    models = join_models(*models)
    return(models,fitnesses_full,hall_of_fame_full,best_fitnesses_full)

"""
from dream5.gene_expression import load_ecoli
from dream5.tf import load_tf_ecoli
from grenadine.Inference.inference import score_links
from grenadine.Inference.regression_predictors import BayesianRidgeScore as BRS
from grenadine.Preprocessing.standard_preprocessing import z_score
from grenadine.Evaluation.evaluation import get_top_tfs_per_tg_from_scores
from grenadine.Model.model import Evolve_optimal_model_tg,Evolve_optimal_model
from grenadine.Model.model import join_models

import numpy as np
import pandas as pd
X = z_score(load_ecoli(),1)
tf = load_tf_ecoli()[1]
scores = score_links(X,BRS,tf_list=tf,progress_bar=True)#,tg_list=["iap","cspA"])
top_tfs = get_top_tfs_per_tg_from_scores(scores,40)

Evolve_optimal_model_tg(X,list(top_tfs.loc["iap"]),'iap',alpha=0.5)

model,fitnesses_full,hall_of_fame_full,best_fitnesses_full = Evolve_optimal_model(X,scores,alpha=0.3)
model.eval
model.tgs_tfs
"""



"""
This module allows to discretize gene expression datasets.
It is mostly based on `scikit-learn`_ library.
Different discretization methods are available :
EWD (equal width, uniform), EFD (equal frequency, quantile), kmeans,
`bikmeans`_ (Li et al., 2010).

.. _scikit-learn:
    https://scikit-learn.org

.. _bikmeans:
    https://www.ncbi.nlm.nih.gov/pubmed/20955620
"""


import pandas as pd
import numpy as np
from math import pow
from sklearn.preprocessing import KBinsDiscretizer


def discretize_genexp(data, method, nb_bins=2, axis=0):
    """
    Discretize data into nb_bins intervals, with specified method, along
    specified axis.

    Args:
        data (pandas.DataFrame or pandas.Series): dataset to discretize
        method (str): method used for discretization, amongst: 'kmeans',
            'bikmeans', 'ewd', 'efd'
        nb_bins (int): (default 2) number of intervals in which to discretize
            data
        axis (int): (default 0) indicates if discretization should be done on
            each column (0) or each line (1) of data. Ignore this parameter if
            method is bikmeans

    Returns:
        pandas.DataFrame or pandas.Series: dataframe or series of discretized
        data, depending on the dimension of passed data

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> data = pd.DataFrame(np.random.randn(3, 5),
                                index=["gene1", "gene2", "gene3"],
                                columns=["c1", "c2", "c3", "c4", "c5"])
        >>> data
                     c1        c2        c3        c4        c5
        gene1  1.764052  0.400157  0.978738  2.240893  1.867558
        gene2 -0.977278  0.950088 -0.151357 -0.103219  0.410599
        gene3  0.144044  1.454274  0.761038  0.121675  0.443863
        >>> discr_data = discretize_genexp(data=data, method='efd')
        >>> discr_data
                c1   c2   c3   c4   c5
        gene1  1.0  0.0  1.0  1.0  1.0
        gene2  0.0  1.0  0.0  0.0  0.0
        gene3  1.0  1.0  1.0  1.0  1.0
    """
    # choose a discretization method
    if method == 'kmeans':
        discr = KBinsDiscretizer(n_bins=nb_bins, encode='ordinal', strategy='kmeans',subsample=None)
    elif method == 'ewd':
        discr = KBinsDiscretizer(n_bins=nb_bins, encode='ordinal', strategy='uniform',subsample=None)
    elif method == 'efd':
        discr = KBinsDiscretizer(n_bins=nb_bins, encode='ordinal', strategy='quantile',subsample=None)
    elif method == 'bikmeans':
        # in this case all discretization process is handled in bikmeans
        discr_data = bikmeans_simple(data, nb_bins)
        return discr_data
    else:
        print("Unvalid option. Methods are: 'kmeans','bikmeans','ewd','efd'")
        return
    # apply discretizer to the data
    if len(data.shape) < 2 :
        # if data is 1D
        discr_data = discr.fit_transform(pd.DataFrame(data).values)
        discr_data = discr_data.flatten()
        if type(data) == type(pd.Series()):
            discr_data = pd.Series(discr_data, index = data.index)
            discr_data.name = data.name
        return(discr_data)
    # if data is 2D
    if axis == 1: # discretize along lines
        discr_data = discr.fit_transform(data.transpose().values)
        discr_data = discr_data.transpose()
    elif axis == 0: # discretize along columns
        discr_data = discr.fit_transform(data.values)
    else:
        print("Unvalid axis argument. 0 for columns (default) and 1 for lines.")
        return
    discr_data = pd.DataFrame(discr_data)
    discr_data.index = data.index
    discr_data.columns = data.columns
    discr_data = discr_data.astype('int64')
    return discr_data


def bikmeans_original(data, nb_bins):
    """
    Discretize data into nb_bins intervals, with method `bikmeans`_, from
    the publication by Li et al, 2010.

    Args:
        data (pandas.DataFrame): dataset to discretize
        nb_bins (int): number of intervals in which to discretize data

    Returns:
        pandas.DataFrame: dataframe of discretized data

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> data = pd.DataFrame(np.random.randn(3, 5),
                                index=["gene1", "gene2", "gene3"],
                                columns=["c1", "c2", "c3", "c4", "c5"])
        >>> data
                     c1        c2        c3        c4        c5
        gene1  1.764052  0.400157  0.978738  2.240893  1.867558
        gene2 -0.977278  0.950088 -0.151357 -0.103219  0.410599
        gene3  0.144044  1.454274  0.761038  0.121675  0.443863
        >>> discr_data = bikmeans_original(data=data, nb_bins=2)
        >>> discr_data
                c1   c2   c3   c4   c5
        gene1  1.0  0.0  0.0  1.0  1.0
        gene2  0.0  1.0  0.0  0.0  0.0
        gene3  0.0  1.0  0.0  0.0  0.0
    """
    km = KBinsDiscretizer(n_bins=nb_bins, encode='ordinal', strategy='kmeans',subsample=None)
    # use kmeans along lines an columns of data separately
    discr1 = km.fit_transform(data.transpose().values).transpose()
    discr2 = km.fit_transform(data.values)
    X = discr1*discr2
    # initalize empty discretized data array
    discr_data = np.zeros(X.shape)
    # going through all values in X and filling in discr_data
    for i,j in np.ndindex(X.shape):
        x = X[i,j]
        k = 0
        while (k < nb_bins):
            if x >= pow(k,2) and x < pow(k+1, 2):
                # discretized value for data[i][j] is k
                discr_data[i][j] = k
                break
            else:
                k += 1
        if k == nb_bins: # error case : no suitable bin was found for x
            discr_data[i][j] = -1
    # cast discretized data into a pandas DataFrame
    discr_data = pd.DataFrame(discr_data)
    discr_data.index = data.index
    discr_data.columns = data.columns
    return discr_data

def bikmeans_simple(data, nb_bins):
    """
    Discretize data into nb_bins intervals, with method `bikmeans`_, simplified.
    From the publication by Li et al, 2010.
    See function bikmeans_original() for the full implementation of bikmeans as
    described in the paper.

    Args:
        data (pandas.DataFrame): dataset to discretize
        nb_bins (int): number of intervals in which to discretize data

    Returns:
        pandas.DataFrame: dataframe of discretized data

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> data = pd.DataFrame(np.random.randn(3, 5),
                                index=["gene1", "gene2", "gene3"],
                                columns=["c1", "c2", "c3", "c4", "c5"])
        >>> data
                     c1        c2        c3        c4        c5
        gene1  1.764052  0.400157  0.978738  2.240893  1.867558
        gene2 -0.977278  0.950088 -0.151357 -0.103219  0.410599
        gene3  0.144044  1.454274  0.761038  0.121675  0.443863
        >>> discr_data = bikmeans_simple(data=data, nb_bins=2)
        >>> discr_data
                c1   c2   c3   c4   c5
        gene1  2.0  1.0  1.0  2.0  2.0
        gene2  1.0  2.0  1.0  1.0  1.0
        gene3  1.0  2.0  1.0  1.0  1.0
    """

    km = KBinsDiscretizer(nb_bins, encode='ordinal', strategy='kmeans')
    # use kmeans along lines an columns of data separately
    discr1 = km.fit_transform(data.transpose().values).transpose()
    discr2 = km.fit_transform(data.values)
    X = (discr1+1)*(discr2+1)
    discr_data = np.floor(np.sqrt(X))
    # cast discretized data into a pandas DataFrame
    discr_data = pd.DataFrame(discr_data)
    discr_data.index = data.index
    discr_data.columns = data.columns
    return discr_data



"""
This module allows to infer co-expression  Gene Regulatory Networks using
gene expression data (RNAseq or Microarray). This module implements severall
inference algorithms based on regression, using `scikit-learn`_.

.. _scikit-learn:
    https://scikit-learn.org
"""
from sklearn.linear_model import BayesianRidge as _sklearn_BayesianRidge
from sklearn.svm import SVR as _sklearn_SVR
from sklearn.linear_model import Lasso as _sklearn_Lasso
from sklearn.linear_model import LassoLars as _sklearn_LassoLars
from sklearn.model_selection import train_test_split as _sklearn_train_test_split
from sklearn.linear_model import Lars as _sklearn_Lars
from sklearn.ensemble import RandomForestRegressor as _sklearn_RandomForestRegressor
from sklearn.ensemble import ExtraTreesRegressor as _sklearn_ExtraTreesRegressor
from sklearn.linear_model import ElasticNetCV as _sklearn_ElasticNetCV
from sklearn.ensemble import GradientBoostingRegressor as _sklearn_GradientBoostingRegressor
from sklearn.ensemble import AdaBoostRegressor as _sklearn_AdaBoostRegressor
from sklearn.ensemble import BaggingRegressor as _sklearn_BaggingRegressor
from pandas import DataFrame
import numpy as np
import pandas as pd
try:
    from sklearn.linear_model import RandomizedLasso as _sklearn_RandomizedLasso
except:
    import sklearn
    print("sklearn.linear_model.RandomizedLasso could not be loaded")
    print('RandomizedLasso is not available in sklearn '+sklearn.__version__)


def GENIE3(X,y,**rf_parameters):
    """
    GENIE3, score predictor function based on `scikit-learn`_
    RandomForestRegressor.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **rf_parameters: Named parameters for the sklearn RandomForestRegressor

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        RandomForestRegressor to the regulatory relationship between the target
        gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = GENIE3(tfs,tg)
        >>> scores
        array([0.11983888, 0.28071399, 0.59944713])
    """
    regressor = _sklearn_RandomForestRegressor(**rf_parameters)
    regressor.fit(X, y)
    scores = regressor.feature_importances_
    return(scores)

def XGENIE3(X,y,**rf_parameters):
    """
    XGENIE3, score predictor function based on `scikit-learn`_
    ExtraTreesRegressor.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **rf_parameters: Named parameters for the sklearn RandomForestRegressor

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        ExtraTreesRegressor to the regulatory relationship between the target
        gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = XGENIE3(tfs,tg)
        >>> scores
        array([0.24905241, 0.43503283, 0.31591477])
    """
    regressor = _sklearn_ExtraTreesRegressor(**rf_parameters)
    regressor.fit(X, y)
    scores = regressor.feature_importances_
    return(scores)

def GRNBoost2(X,y,**boost_parameters):
    """
    GRNBoost2 score predictor based on `scikit-learn`_
    GradientBoostingRegressor.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **boost_parameters: Named parameters for GradientBoostingRegressor

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        GradientBoostingRegressor to the regulatory relationship between the
        target gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = GRNBoost2(tfs,tg)
        >>> scores
        array([0.83904506, 0.01783977, 0.14311517])
    """
    regressor = _sklearn_GradientBoostingRegressor(**boost_parameters)
    regressor.fit(X, y)
    scores = regressor.feature_importances_
    return(scores)

def BayesianRidgeScore(X,y,**brr_parameters):
    """
    Score predictor based on `scikit-learn`_ BayesianRidge regression.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **brr_parameters: Named parameters for sklearn BayesianRidge regression

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        sklearn BayesianRidge regressor to the regulatory relationship between
        the target gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = BayesianRidgeScore(tfs,tg)
        >>> scores
        array([1.32082000e-03, 6.24177371e-05, 3.32319918e-04])
    """
    regressor = _sklearn_BayesianRidge(**brr_parameters)
    regressor.fit(X, y)
    scores = np.abs(regressor.coef_)
    return(scores)

def SVR_score(X,y,**svr_parameters):
    """
    Score predictor based on `scikit-learn`_ SVR (Support Vector Regression).

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **svr_parameters: Named parameters for sklearn SVR regression

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        sklearn SVR regressor to the regulatory relationship between
        the target gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = SVR_score(tfs,tg)
        >>> scores
        array([[-0.38156814,  0.28128811, -1.0230867 ]])
    """
    svr_parameters["kernel"] = 'linear'
    regressor = _sklearn_SVR(**svr_parameters)
    regressor.fit(X, y)
    scores = np.abs(regressor.coef_)
    return(scores[0])

def Lasso_score(X,y,**l1_parameters):
    """
    Score predictor based on `scikit-learn`_ Lasso regression.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **l1_parameters: Named parameters for sklearn Lasso regression

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        sklearn Lasso regressor to the regulatory relationship between
        the target gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = Lasso_score(tfs,tg, alpha=0.01)
        >>> scores
        array([0.13825495, 0.94939204, 0.19118214])
    """
    regressor = _sklearn_Lasso(**l1_parameters)
    regressor.fit(X, y)
    scores = np.abs(regressor.coef_)
    return(scores)

def LassoLars_score(X,y,**l1_parameters):
    """
    Score predictor based on `scikit-learn`_ LassoLars regression.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **l1_parameters: Named parameters for sklearn Lasso regression

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        sklearn LassoLars regressor to the regulatory relationship between the
        target gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = LassoLars_score(tfs,tg, alpha=0.01)
        >>> scores
        array([0.12179406, 0.92205553, 0.15503451])
    """
    regressor = _sklearn_LassoLars(**l1_parameters)
    regressor.fit(X, y)
    scores = np.abs(regressor.coef_)
    return(scores)

def stability_randomizedlasso(X,y,**rl_parameters):
    """
    Score predictor based on `scikit-learn`_ randomizedlasso stability selection.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **rl_parameters: Named parameters for sklearn randomizedlasso
    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        sklearn randomizedlasso stability selection to the regulatory
        relationship between the target gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = stability_randomizedlasso(tfs,tg)
        >>> scores
        array([0.11 , 0.17 , 0.085])
    """
    regressor = _sklearn_RandomizedLasso(**rl_parameters)
    regressor.fit(X,y)
    scores = np.abs(regressor.scores_)
    return(scores)

def TIGRESS(X,
            y,
            nsplit=100,
            nstepsLARS=5,
            alpha=0.4,
            scoring="area"):
    """
    TIGRESS score predictor based on stability selection.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        nsplit (int): number of splits applied,
            i.e., randomization tests, the highest the best
        nstepsLARS (int): number of steps of LARS algorithm,
            i.e., number of non zero coefficients to keep (Lars parameter)
        alpha: Noise multiplier coefficient,
            Each transcription factor expression is multiplied by a
            random variable $\in [\alpha,1]$
        scoring (str): option used to score each possible link
            only "area" and "max" options are available

    Returns:
        numpy.array: co-regulation scores

        The i-th element of the score array represents the score assigned by the
        sklearn randomizedlasso stability selection to the regulatory
        relationship between the target gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                               index =["c1","c2","c3","c4","c5"],
                               columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = TIGRESS(tfs,tg)
        >>> scores
        array([349.   , 312.875, 588.125])
    """
    n,p = X.shape
    halfsize = int(n/2)
    if nstepsLARS > p:
        nstepsLARS = p-1
    freq = np.zeros((p, nstepsLARS))
    i = 0
    while i < nsplit:
        # Randomly reweight each variable (TF expression)
        random_perturbation = np.random.uniform(low=alpha, high=1.0, size=p)
        X *= random_perturbation
        # Randomly split the sample in two sets
        X_1,X_2,y_1,y_2 = _sklearn_train_test_split(X,y,test_size=halfsize, shuffle=True)
        for X_i,y_i in [[X_1, y_1],[X_2,y_2]]:
            if y_i.std() > 0:
                # run LARS on each subsample and collect variables are selected
                lars = _sklearn_Lars(normalize=False, n_nonzero_coefs=nstepsLARS)
                lars.fit(X_i,y_i)
                # collect the presence of the coefficients along the path
                path = lars.coef_path_
                if path.shape[1] < nstepsLARS+1:
                    path_add = np.tile(path[:,-1],(nstepsLARS+1 - path.shape[1],1)).T
                    path = np.hstack((path,path_add))
                freq += np.abs(np.sign(path[:,1:]))
                i += 1
        X /= random_perturbation
    # normalize frequence in [0,1] to get stability curves
    freq /= 2*halfsize
    if (scoring=="area"):
        score = np.cumsum(freq,axis=1)/np.arange(1,nstepsLARS+1,1)
    if (scoring=="max"):
        score = np.maximum.accumulate(freq,axis=1)
    return(score[:,nstepsLARS - 1])

def AdaBoost_regressor (X, y,  **adab_parameters):
    """
    AdaBoost regressor, score predictor function based on `scikit-learn`_
    AdaBoostRegressor.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **adab_parameters: Named parameters for the sklearn AdaBoostRegressor

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        AdaBoostRegressor to the regulatory relationship between the target
        gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                       index =["c1","c2","c3","c4","c5"],
                       columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = AdaBoost_regressor(tfs,tg)
        >>> scores
        array([0.32978247, 0.3617295 , 0.28896647])
    """
    regressor = _sklearn_AdaBoostRegressor(**adab_parameters)
    regressor.fit(X, y)
    scores = regressor.feature_importances_
    return scores

def Elastica(X, y, **elastica_parameters):
    """
    ElasticNetCV regressor, score predictor function based on `scikit-learn`_
    ElasticNetCV.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **elastica_parameters: Named parameters for the sklearn ElasticNetCV

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the score assigned by the
        AdaBoostRegressor to the regulatory relationship between the target
        gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                       index =["c1","c2","c3","c4","c5"],
                       columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> scores = Elastica(tfs,tg)
        >>> scores
        array([0.05512459, 0.34453337, 0.        ])
    """
    regressor = _sklearn_ElasticNetCV(**elastica_parameters)
    regressor.fit(X, y)
    scores = np.abs(regressor.coef_)
    return scores


def bagging_regressor(X, y, **bagging_parameters):
    """
    Apply the bagging technique to a regression algorithm, based on
    `scikit-learn`_ BaggingRegressor.

    Args:
        X (pandas.DataFrame): Transcriptor factor gene expressions where rows
            are experimental conditions and columns are transcription factors
        y (pandas.Series): Target gene expression vector where rows are
            experimental conditions
        **adab_parameters: Named parameters for the sklearn AdaBoostRegressor

    Returns:
        numpy.array: co-regulation scores.

        The i-th element of the score array represents the average score
        assigned by the Base Regressor to the regulatory relationship
        between the target gene and transcription factor i.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> from sklearn.svm import SVR
        >>> np.random.seed(0)
        >>> svr = SVR(kernel="linear")
        >>> tfs = pd.DataFrame(np.random.randn(5,3),
                       index =["c1","c2","c3","c4","c5"],
                       columns=["tf1","tf2","tf3"])
        >>> tg = pd.Series(np.random.randn(5),index=["c1","c2","c3","c4","c5"])
        >>> bagging_parameters = {"base_estimator":svr,
                                  "n_estimators":100,
                                  "max_samples":0.7}
        >>> scores = bagging_regressor(tfs,tg,**bagging_parameters)
        >>> scores
        array([0.32978247, 0.3617295 , 0.28896647])
    """
    def get_score(classifier):
        scores=None
        if hasattr(classifier, 'feature_importances_'):
            scores = DataFrame(feature_importances).mean(axis=0)
        elif hasattr(classifier,"coef_"):
            scores = np.abs(classifier.coef_).flatten()
        return(scores)
    bc = _sklearn_BaggingRegressor(**bagging_parameters)
    bc.fit(X,y)
    coefs = [get_score(e) for e in bc.estimators_]
    coefs = [pd.Series(c,index=bc.estimators_features_[i]) for i,c in enumerate(coefs) if c is not None]
    if len(coefs):
        scores = pd.Series(np.random.randn(bc.n_features_)*1e-5)
        scores = scores + pd.DataFrame(coefs).mean(axis=0)
    else:
        print("Base regressor has not a feature_importance of coef_ attribute")
        return(0)
    scores = scores.sort_index()
    return(scores)



# -*- coding: utf-8 -*-
"""
This module allows to infer co-expression  Gene Regulatory Networks using
gene expression data (RNAseq or Microarray).
"""
from tqdm.auto import tqdm
from sklearn.model_selection import GridSearchCV
from grenadine.Preprocessing.standard_preprocessing import z_score
import pandas as pd
import numpy as np
import warnings
import os

class gene_model:
    """
    Make a gene-level self expressive optimized model

    Args:
        params_gridsearch (dict): parameters for sklearn GridSearchCV to perform 
        a meta-parameter optimization for the gene-level self-expressive model

    Examples:
        >>> from sklearn.tree import DecisionTreeClassifier
        >>> from sklearn.model_selection import StratifiedKFold
        >>> model = DecisionTreeClassifier(min_samples_leaf=3)
        >>> params = {"min_impurity_decrease":[1e-2,5e-2,1e-1,5e-1],
                      "ccp_alpha":[1e-1,5e-2,1e-2,5e-3,1e-3],
                     "criterion":["gini","entropy",]
                    }
        >>> params_gridsearch = {"estimator":model,
                                 "param_grid":params,
                                 "return_train_score":True,
                                 "scoring":["balanced_accuracy",
                                             "accuracy",
                                             "f1_weighted",
                                             "f1_macro",
                                             "recall_weighted",
                                             "recall_macro",
                                             "roc_auc_ovr_weighted"],
                                 "refit":"roc_auc_ovr_weighted",
                                 "cv":StratifiedKFold(5),
                                 "n_jobs":-1,
                                }
        >>> gm = gene_model(params_gridsearch)


    """
    def __init__(self,params_gridsearch):
        self.params_gridsearch = params_gridsearch
        self.gscv = GridSearchCV(**params_gridsearch)
    
    def fit(self,X,y):
        """
        Fit the self expressive gene-level model

        Args:
            X (pandas.DataFrame): Transcription factor gene expressions (discretized
                or not) where rows are experimental conditions and columns are transcription factors
            y (pandas.Series): Target gene expression vector (discretized) where
                rows are experimental conditions

        Examples:
            >>> from sklearn.tree import DecisionTreeClassifier
            >>> from sklearn.model_selection import StratifiedKFold
            >>> model = DecisionTreeClassifier(min_samples_leaf=3,random_state=0)
            >>> params = {"min_impurity_decrease":[1e-2,5e-2,1e-1,5e-1],
                          "ccp_alpha":[1e-1,5e-2,1e-2,5e-3,1e-3],
                         "criterion":["gini","entropy",]
                        }
            >>> params_gridsearch = {"estimator":model,
                                     "param_grid":params,
                                     "return_train_score":True,
                                     "scoring":["balanced_accuracy",
                                                 "accuracy",
                                                 "f1_weighted",
                                                 "f1_macro",
                                                 "recall_weighted",
                                                 "recall_macro",
                                                 "roc_auc_ovr_weighted"],
                                     "refit":"roc_auc_ovr_weighted",
                                     "cv":StratifiedKFold(5),
                                     "n_jobs":-1,
                                    }
            >>> gm = gene_model(params_gridsearch)
            >>> import pandas as pd
            >>> import numpy as np
            >>> np.random.seed(0)
            >>> X = pd.DataFrame(np.random.randn(50,3),
                                 columns=["tf1","tf2","tf3"])
            >>> tg = pd.Series(np.random.randint(0,3,size=50))
            >>> gm.fit(X,tg)
        """
        self.gscv.fit(X,y)
    
    def regulators_importance(self):
        """
        Compute the feature importance of regulators using the feature_importance_ or the score_ attribute from sklearn model

        Args:
            X (pandas.DataFrame): Transcription factor gene expressions (discretized
                or not) where rows are experimental conditions and columns are transcription factors
            y (pandas.Series): Target gene expression vector (discretized) where
                rows are experimental conditions

        Examples:
            >>> from sklearn.tree import DecisionTreeClassifier
            >>> from sklearn.model_selection import StratifiedKFold
            >>> model = DecisionTreeClassifier(min_samples_leaf=3,random_state=0)
            >>> params = {"min_impurity_decrease":[1e-2,5e-2,1e-1,5e-1],
                          "ccp_alpha":[1e-1,5e-2,1e-2,5e-3,1e-3],
                         "criterion":["gini","entropy",]
                        }
            >>> params_gridsearch = {"estimator":model,
                                     "param_grid":params,
                                     "return_train_score":True,
                                     "scoring":["balanced_accuracy",
                                                 "accuracy",
                                                 "f1_weighted",
                                                 "f1_macro",
                                                 "recall_weighted",
                                                 "recall_macro",
                                                 "roc_auc_ovr_weighted"],
                                     "refit":"roc_auc_ovr_weighted",
                                     "cv":StratifiedKFold(5),
                                     "n_jobs":-1,
                                    }
            >>> gm = gene_model(params_gridsearch)
            >>> import pandas as pd
            >>> import numpy as np
            >>> np.random.seed(0)
            >>> X = pd.DataFrame(np.random.randn(50,3),
                                 columns=["tf1","tf2","tf3"])
            >>> tg = pd.Series(np.random.randint(0,3,size=50))
            >>> gm.fit(X,tg)
            >>> gm.regulators_importance()
            tf1    0.407930
            tf2    0.166075
            tf3    0.425995
            dtype: float64
        """
        features = self.gscv.best_estimator_.feature_names_in_
        importance = None
        if "feature_importances_" in self.gscv.best_estimator_.__dir__():
            importance = self.gscv.best_estimator_.feature_importances_
        elif "coef_" in self.gscv.best_estimator_.__dir__():
            importance = self.gscv.best_estimator_.coef_.mean(axis=0)
        if importance is not None:
            return pd.Series(importance, index=features)
    
    def get_scores(self):
        """
        Return the scores for the best self-expressive gene-level model 

        Args:
            X (pandas.DataFrame): Transcription factor gene expressions (discretized
                or not) where rows are experimental conditions and columns are transcription factors
            y (pandas.Series): Target gene expression vector (discretized) where
                rows are experimental conditions

        Examples:
            >>> from sklearn.tree import DecisionTreeClassifier
            >>> from sklearn.model_selection import StratifiedKFold
            >>> model = DecisionTreeClassifier(min_samples_leaf=3,random_state=0)
            >>> params = {"min_impurity_decrease":[1e-2,5e-2,1e-1,5e-1],
                          "ccp_alpha":[1e-1,5e-2,1e-2,5e-3,1e-3],
                         "criterion":["gini","entropy",]
                        }
            >>> params_gridsearch = {"estimator":model,
                                     "param_grid":params,
                                     "return_train_score":True,
                                     "scoring":["balanced_accuracy",
                                                 "accuracy",
                                                 "f1_weighted",
                                                 "f1_macro",
                                                 "recall_weighted",
                                                 "recall_macro",
                                                 "roc_auc_ovr_weighted"],
                                     "refit":"roc_auc_ovr_weighted",
                                     "cv":StratifiedKFold(5),
                                     "n_jobs":-1,
                                    }
            >>> gm = gene_model(params_gridsearch)
            >>> import pandas as pd
            >>> import numpy as np
            >>> np.random.seed(0)
            >>> X = pd.DataFrame(np.random.randn(50,3),
                                 columns=["tf1","tf2","tf3"])
            >>> tg = pd.Series(np.random.randint(0,3,size=50))
            >>> gm.fit(X,tg)
            >>> gm.get_scores()
            mean_fit_time                        0.001037
            std_fit_time                         0.000429
            mean_score_time                      0.003107
            std_score_time                        0.00008
            param_ccp_alpha                          0.01
                                                   ...   
            split2_train_roc_auc_ovr_weighted    0.882115
            split3_train_roc_auc_ovr_weighted    0.894333
            split4_train_roc_auc_ovr_weighted    0.896417
            mean_train_roc_auc_ovr_weighted      0.905028
            std_train_roc_auc_ovr_weighted       0.019257
            Name: 16, Length: 113, dtype: object
        """
        best = self.gscv.best_index_
        return pd.DataFrame(self.gscv.cv_results_).iloc[best,:]
            
class gxn_model:
    """
    Make a genme-level self expressive optimized model

    Args:
        params_gridsearch (dict): parameters for sklearn GridSearchCV to perform 
        a meta-parameter optimization for each gene-level self-expressive model
        verbose (bool): print extra information 

    Examples:
        >>> from sklearn.tree import DecisionTreeClassifier
        >>> from sklearn.model_selection import StratifiedKFold
        >>> model = DecisionTreeClassifier(min_samples_leaf=3)
        >>> params = {"min_impurity_decrease":[1e-2,5e-2,1e-1,5e-1],
                      "ccp_alpha":[1e-1,5e-2,1e-2,5e-3,1e-3],
                     "criterion":["gini","entropy",]
                    }
        >>> params_gridsearch = {"estimator":model,
                                 "param_grid":params,
                                 "return_train_score":True,
                                 "scoring":["balanced_accuracy",
                                             "accuracy",
                                             "f1_weighted",
                                             "f1_macro",
                                             "recall_weighted",
                                             "recall_macro",
                                             "roc_auc_ovr_weighted"],
                                 "refit":"roc_auc_ovr_weighted",
                                 "cv":StratifiedKFold(5),
                                 "n_jobs":-1,
                                }
        >>> gm = gene_model(params_gridsearch)


    """
    def __init__(self,params_gridsearch,verbose=True):
        self.params_gridsearch = params_gridsearch
        self.verbose = verbose
        
    def fit(self,X,Y,regulators,to_inspect=None):  
        """
        Fit the self expressive gene regulatory model

        Args:
            X (pandas.DataFrame): Predictor gene expression matrix where 
                rows are experimental conditions and columns are genes
            Y (pandas.DataFrame): Target gene expression matrix (discretized if classification model used) 
                where rows are experimental conditions
            regulators (list): list of columns in X used as predictors (i.e., regulators)
            to_inspect (list): list of columns in Y to model, if None all genes in Y are analyzed

        Examples:
            >>> from sklearn.tree import DecisionTreeClassifier
            >>> from sklearn.model_selection import StratifiedKFold
            >>> model = DecisionTreeClassifier(min_samples_leaf=3,random_state=0)
            >>> params = {"min_impurity_decrease":[1e-2,5e-2,1e-1,5e-1],
                          "ccp_alpha":[1e-1,5e-2,1e-2,5e-3,1e-3],
                         "criterion":["gini","entropy",]
                        }
            >>> params_gridsearch = {"estimator":model,
                                     "param_grid":params,
                                     "return_train_score":True,
                                     "scoring":["balanced_accuracy",
                                                 "accuracy",
                                                 "f1_weighted",
                                                 "f1_macro",
                                                 "recall_weighted",
                                                 "recall_macro",
                                                 "roc_auc_ovr_weighted"],
                                     "refit":"roc_auc_ovr_weighted",
                                     "cv":StratifiedKFold(5),
                                     "n_jobs":-1,
                                    }
            >>> gxn = gxn_model(params_gridsearch)
            >>> import pandas as pd
            >>> import numpy as np
            >>> np.random.seed(0)
            >>> X = pd.DataFrame(np.random.randn(50,3),
                                 columns=["tf1","tf2","tf3"])
            >>> from grenadine.Preprocessing.discretization import discretize_genexp
            >>> Y = discretize_genexp(X,method="kmeans",nb_bins=2,axis=0)
            >>> regulators = ["tf1","tf2"]
            >>> gxn.fit(X,Y,regulators)
        """
        self.gene_models_ = {}
        if to_inspect is None:
            to_inspect = Y.columns
        for i,gene in tqdm(enumerate(to_inspect)):
            if i%100 == 0:
                _ = os.write(1,bytes(f"processed: {i} genes \n",'utf-8'))
            self.gene_models_[gene] = gene_model(self.params_gridsearch)
            X_train = X[[e for e in regulators if e!=gene]]
            y_train = Y[gene]
            self.gene_models_[gene].fit(X_train,y_train)
            
    def transform(self):
        """
        Build the gene self expressive network coefficients

        Returns:
            pandas.DataFrame: self expressive network coefficients matrix.
            Rows are target genes and columns are regulators.
            The value at row i and column j represents the feature importance score
            assigned by the self-expressive model to the regulatory relationship 
            between target gene i and regulator j.   

        Examples:
            >>> from sklearn.tree import DecisionTreeClassifier
            >>> from sklearn.model_selection import StratifiedKFold
            >>> model = DecisionTreeClassifier(min_samples_leaf=3,random_state=0)
            >>> params = {"min_impurity_decrease":[1e-2,5e-2,1e-1,5e-1],
                          "ccp_alpha":[1e-1,5e-2,1e-2,5e-3,1e-3],
                         "criterion":["gini","entropy",]
                        }
            >>> params_gridsearch = {"estimator":model,
                                     "param_grid":params,
                                     "return_train_score":True,
                                     "scoring":["balanced_accuracy",
                                                 "accuracy",
                                                 "f1_weighted",
                                                 "f1_macro",
                                                 "recall_weighted",
                                                 "recall_macro",
                                                 "roc_auc_ovr_weighted"],
                                     "refit":"roc_auc_ovr_weighted",
                                     "cv":StratifiedKFold(5),
                                     "n_jobs":-1,
                                    }
            >>> gxn = gxn_model(params_gridsearch)
            >>> import pandas as pd
            >>> import numpy as np
            >>> np.random.seed(0)
            >>> X = pd.DataFrame(np.random.randn(50,3),
                                 columns=["tf1","tf2","tf3"])
            >>> from grenadine.Preprocessing.discretization import discretize_genexp
            >>> Y = discretize_genexp(X,method="kmeans",nb_bins=2,axis=0)
            >>> regulators = ["tf1","tf2"]
            >>> gxn.fit(X,Y,regulators)
            >>> gxn.transform()
        """
        self.C_ = {m: self.gene_models_[m].regulators_importance() for m in self.gene_models_} 
        self.C_ = pd.DataFrame(self.C_)
        self.C_ = self.C_.T
        self.C_.fillna(0,inplace=True)
        return self.C_
    
    def score(self):
        """
        Return the scores for the best self-expressive model for each gene

        Returns:
            pandas.DataFrame: Evaluation scores for each gene, rows represent genes and columns metrics

        Examples:
            >>> from sklearn.tree import DecisionTreeClassifier
            >>> from sklearn.model_selection import StratifiedKFold
            >>> model = DecisionTreeClassifier(min_samples_leaf=3,random_state=0)
            >>> params = {"min_impurity_decrease":[1e-2,5e-2,1e-1,5e-1],
                          "ccp_alpha":[1e-1,5e-2,1e-2,5e-3,1e-3],
                         "criterion":["gini","entropy",]
                        }
            >>> params_gridsearch = {"estimator":model,
                                     "param_grid":params,
                                     "return_train_score":True,
                                     "scoring":["balanced_accuracy",
                                                 "accuracy",
                                                 "f1_weighted",
                                                 "f1_macro",
                                                 "recall_weighted",
                                                 "recall_macro",
                                                 "roc_auc_ovr_weighted"],
                                     "refit":"roc_auc_ovr_weighted",
                                     "cv":StratifiedKFold(5),
                                     "n_jobs":-1,
                                    }
            >>> gxn = gxn_model(params_gridsearch)
            >>> import pandas as pd
            >>> import numpy as np
            >>> np.random.seed(0)
            >>> X = pd.DataFrame(np.random.randn(50,3),
                                 columns=["tf1","tf2","tf3"])
            >>> from grenadine.Preprocessing.discretization import discretize_genexp
            >>> Y = discretize_genexp(X,method="kmeans",nb_bins=2,axis=0)
            >>> regulators = ["tf1","tf2"]
            >>> gxn.fit(X,Y,regulators)
            >>> gxn.score()
            
        """
        self.scores_ = pd.DataFrame({m: self.gene_models_[m].get_scores() for m in self.gene_models_})
        self.scores_ = self.scores_.T
        self.scores_.fillna(0,inplace=True)
        return self.scores_
    
    def purify_model(self,score_condition,inplace=False):
        """
        Return a gxn_model purified keeping only genes with score following a given criteria

        Returns:
            gxn_model : purified model

        Examples:
            >>> from sklearn.tree import DecisionTreeClassifier
            >>> from sklearn.model_selection import StratifiedKFold
            >>> model = DecisionTreeClassifier(min_samples_leaf=3,random_state=0)
            >>> params = {"min_impurity_decrease":[1e-2,5e-2,1e-1,5e-1],
                          "ccp_alpha":[1e-1,5e-2,1e-2,5e-3,1e-3],
                         "criterion":["gini","entropy",]
                        }
            >>> params_gridsearch = {"estimator":model,
                                     "param_grid":params,
                                     "return_train_score":True,
                                     "scoring":["balanced_accuracy",
                                                 "accuracy",
                                                 "f1_weighted",
                                                 "f1_macro",
                                                 "recall_weighted",
                                                 "recall_macro",
                                                 "roc_auc_ovr_weighted"],
                                     "refit":"roc_auc_ovr_weighted",
                                     "cv":StratifiedKFold(5),
                                     "n_jobs":-1,
                                    }
            >>> gxn = gxn_model(params_gridsearch)
            >>> import pandas as pd
            >>> import numpy as np
            >>> np.random.seed(0)
            >>> X = pd.DataFrame(np.random.randn(50,3),
                                 columns=["tf1","tf2","tf3"])
            >>> from grenadine.Preprocessing.discretization import discretize_genexp
            >>> Y = discretize_genexp(X,method="kmeans",nb_bins=2,axis=0)
            >>> regulators = ["tf1","tf2"]
            >>> gxn.fit(X,Y,regulators)
            >>> def score_criteria(score):
                    score["mean_train_roc_auc_ovr_weighted"]>0.6 and score["mean_test_average_precision"]>0.6
            >>> gxn.purify_model(score_criteria)

        """
        mask = [i for i in self.scores_.index if score_condition(self.scores_.loc[i])]
        gene_models_ = {i:self.gene_models_[i] for i in mask}
        C_ = self.C_.loc[mask]
        scores_ = self.scores_.loc[mask]
        if inplace:
            self.gene_models_ = gene_models_
            self.C_ = C_
            self.scores_ = scores_
            return self
        else:
            purified_model = gxn_model(self.params_gridsearch)
            purified_model.gene_models_ = gene_models_
            purified_model.C_ = C_
            purified_model.scores_ = scores_
            return purified_model

def score_links(gene_expression_matrix,
                score_predictor,
                tf_list = None,
                tg_list = None,
                normalize = False,
                discr_method=None,
                progress_bar=False,
                **predictor_parameters):
    """
    Scores transcription factors-target gene co-expressions using a predictor.

    Args:
        gene_expression_matrix (pandas.DataFrame):  gene expression matrix where
            rows are genes and  columns ares samples (conditions).
            The value at row i and column j represents the expression of gene i
            in condition j.
        score_predictor (function): function that receives a pandas.DataFrame X
            containing the transcriptor factor expressions and a pandas.Series
            y containing the expression of a target gene, and scores the
            co-expression level between each transcription factor and the target
            gene.
        tf_list (list or numpy.array): list of transcription factors ids.
        tg_list (list or numpy.array): list of target genes ids.
        normalize (boolean): If True the gene expression of genes is z-scored
        discr_method : discretization method to use, if discretization of target
            gene expression is desired
        progress_bar: bool, if true include progress bar
        **predictor_parameters: Named parameters for the score predictor

    Returns:
        pandas.DataFrame: co-regulation score matrix.

        Rows are target genes and columns are transcription factors.
        The value at row i and column j represents the score assigned by the
        score_predictor to the regulatory relationship between target gene i
        and transcription factor j.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> data = pd.DataFrame(np.random.randn(5, 5),
                            index=["gene1", "gene2", "gene3", "gene4", "gene5"],
                            columns=["c1", "c2", "c3", "c4", "c5"])
        >>> tf_list = ["gene1", "gene2", "gene5"]

        >>> # Example with a regression method
        >>> from grenadine.Inference.regression_predictors import GENIE3
        >>> scores1 = score_links(gene_expression_matrix=data,
                                  score_predictor=GENIE3,
                                  tf_list=tf_list)
        >>> scores1
                  gene2     gene5     gene1
        gene1  0.484081  0.515919       NaN
        gene2       NaN  0.653471  0.346529
        gene3  0.245136  0.301229  0.453634
        gene4  0.309982  0.306964  0.383054
        gene5  0.529839       NaN  0.470161

        >>> # Example with a classification method
        >>> from grenadine.Inference.classification_predictors import RF_classifier_score
        >>> from grenadine.Preprocessing.discretization import discretize_genexp
        >>> discr_method = lambda X: discretize_genexp (X, "efd", 5, axis=1)
        >>> scores2 = score_links(gene_expression_matrix=data,
                                        score_predictor=RF_classifier_score,
                                        tf_list=tf_list,
                                        discr_method=discr_method)
        >>> scores2
                  gene2     gene5     gene1
        gene1  0.512659  0.487341       NaN
        gene2       NaN  0.463122  0.536878
        gene3  0.368175  0.317341  0.314484
        gene4  0.302738  0.346799  0.350463
        gene5  0.524815       NaN  0.475185

    """

    warnings.warn("score_links is deprecated and no longer mantained, please build a gxn_model instead", DeprecationWarning)
    gene_expression_matrix = gene_expression_matrix.T
    # Set the list of TFs and TGs if necessary
    if tg_list is None:
        tg_list = gene_expression_matrix.columns
    if tf_list is None:
        tf_list = gene_expression_matrix.columns
    tf_list_present = set(gene_expression_matrix.columns).intersection(tf_list)
    tg_list_present = list(set(gene_expression_matrix.columns).intersection(tg_list))
    if not len(tf_list_present):
        Exception('None of the tfs in '+str(tf_list)+\
        " is present in the gene_expression_matrix genes list"+\
        str(gene_expression_matrix.columns))
    if not len(tg_list_present):
        Exception('None of the tgs in '+str(tg_list)+\
        " is present in the gene_expression_matrix genes list"+\
        str(gene_expression_matrix.columns))
    tg_list_present.sort()
    #  Normalize expression data for each gene
    if normalize:
        gene_expression_matrix = z_score(gene_expression_matrix,axis=0)
    # compute tf scores for each gene
    scores_tf_per_gene = []
    for gene in tqdm(tg_list_present,disable=not progress_bar):
        # Exclude the current gene from the tfs list
        tfs2test = list(tf_list_present.difference(set([gene])))
        tfs2test.sort()
        X = gene_expression_matrix[tfs2test].values
        y = gene_expression_matrix[gene].values
        if discr_method is not None :
            y = discr_method(y)
        if len(np.unique(y)) <= 1:
            # handle the case when only one class was detected in y
            score = np.zeros(len(tfs2test))
        else:
            score = score_predictor(X, y, **predictor_parameters)
        # Get the features importance (score for each TF -> Gene)
        scores = pd.Series(score)
        scores.index = tfs2test
        scores_tf_per_gene.append(scores)
    df_results = pd.DataFrame(scores_tf_per_gene, index=tg_list_present)
    return(df_results)



def ensemble_score_links(score_links_matrices, score_links_weights=None):
    """
    Makes an ensemble co-regulation score matrix from a list of co-regulation
    score matrices obtained using different methods, and possibly a list of
    weights for each method

    Args:
        score_links_matrices (list): list of co-regulation score matrices (pandas
            DataFrames)
        score_links_weights (list): list of weights for each method (the higher
            the more confidence on the method). If no value is provided each
            method as a unitary weight

    Returns:
        pandas.DataFrame: co-regulation score matrix.

        Rows are target genes and columns are transcription factors.
        The value at row i and column j represents the score assigned by the
        score_predictor to the regulatory relationship between target gene i
        and transcription factor j.
    """
    if score_links_weights is None:
        score_links_weights = np.asarray([1 for s in score_links_matrices])

    for i,score in enumerate(score_links_matrices):
        score_links_matrices[i] -= score_links_matrices[i].values.flatten().mean()
        score_links_matrices[i] /= score_links_matrices[i].values.flatten().std()
        score_links_matrices[i] *= score_links_weights[i]
        if i == 0:
            score_links = score_links_matrices[i]
        else:
            score_links += score_links_matrices[i]
    return score_links/score_links_weights.sum()


def clean_nan_inf_scores(scores):
    """
    Replaces nan and -inf scores by the (minimum_score - 1), and inf scores by
    (maximum_score + 1)

    Args:
        scores (pandas.DataFrame): co-regulation score matrix.
        Rows are target genes and columns are transcription factors.
        The value at row i and column j represents the score assigned by the
        score_predictor to the regulatory relationship between target gene i
        and transcription factor j.

    Returns:
        pandas.DataFrame: co-regulation score matrix.

        Rows are target genes and columns are transcription factors.
        The value at row i and column j represents the score assigned by the
        score_predictor to the regulatory relationship between target gene i
        and transcription factor j.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> data = pd.DataFrame(np.random.randn(5, 5),
                            index=["gene1", "gene2", "gene3", "gene4", "gene5"],
                            columns=["c1", "c2", "c3", "c4", "c5"])
        >>> tf_list = ["gene1", "gene2", "gene5"]

        >>> # Example with a regression method
        >>> from grenadine.Inference.regression_predictors import GENIE3
        >>> scores1 = score_links(gene_expression_matrix=data,
                                  score_predictor=GENIE3,
                                  tf_list=tf_list)
        >>> scores1
                  gene2     gene5     gene1
        gene1  0.484081  0.515919       NaN
        gene2       NaN  0.653471  0.346529
        gene3  0.245136  0.301229  0.453634
        gene4  0.309982  0.306964  0.383054
        gene5  0.529839       NaN  0.470161
        >>> clean_nan_inf_scores(scores1)
                  gene2     gene5     gene1
        gene1  0.484081  0.515919  0.245126
        gene2  0.245126  0.653471  0.346529
        gene3  0.245136  0.301229  0.453634
        gene4  0.309982  0.306964  0.383054
        gene5  0.529839  0.245126  0.470161

    """
    # replacing NaN scores with minimum score - epsilon
    epsilon = 1e-10
    nan_scores = np.isnan(scores)
    min_score =  scores[np.logical_not(nan_scores)].min().min()
    scores[nan_scores] = min_score-epsilon
    # replacing infinite scores with maximum score + epsilon
    finite_scores = np.isfinite(scores)
    max_score =  scores[finite_scores].max().max()
    positive_inf = np.logical_and(np.logical_not(finite_scores),finite_scores>0)
    negative_inf = np.logical_and(np.logical_not(finite_scores),finite_scores<0)
    scores[positive_inf] = max_score+epsilon
    scores[positive_inf] = min_score-epsilon
    return scores


def rank_GRN(coexpression_scores_matrix, take_abs_score=False, clean_scores=True, pyscenic_format=False):
    """
    Ranks the co-regulation scores between transcription factors and target genes.

    Args:
        coexpression_scores_matrix (pandas.DataFrame):co-expression score matrix
            where rows are target genes and columns are transcription factors.
            The value at row i and column j represents the score assigned by a
            score_predictor to the regulatory relationship between target gene i
            and transcription factor j.
        take_abs_score (bool): take the absolute value of the score instead of
            taking scores themselves
    Returns:
        pandas.DataFrame: ranking matrix.

        A ranking matrix contains a row for each possible regulatory link, it
        also contains 4 columns, namely the rank, the score, the transcription
        factor id, and the target gene id.

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> data = pd.DataFrame(np.random.randn(3, 2),
                            index=["gene1", "gene2", "gene3"],
                            columns=["gene1", "gene3"])
        >>> # scores associated to self loops are set to nan
        >>> data.iloc[0,0]=np.nan
        >>> data.iloc[2,1]=np.nan
        >>> ranking_matrix = rank_GRN(data)
        >>> ranking_matrix
                     rank     score     TF     TG
        gene3_gene2   1.0  2.240893  gene3  gene2
        gene1_gene3   2.0  1.867558  gene1  gene3
        gene1_gene2   3.0  0.978738  gene1  gene2
        gene3_gene1   4.0  0.400157  gene3  gene1

    """
    epsilon = 1e-10
    rand_values = np.random.rand(*coexpression_scores_matrix.shape)*epsilon
    coexpression_scores_matrix += rand_values
    if clean_scores:
        coexpression_scores_matrix = clean_nan_inf_scores(coexpression_scores_matrix)
    if take_abs_score:
        coexpression_scores_matrix = np.abs(coexpression_scores_matrix)
    coexpression_unstack = coexpression_scores_matrix.unstack(level=0)
    ranking = pd.DataFrame()
    ranking["rank"] = coexpression_unstack.rank(method="dense",
                                                ascending=False,
                                                na_option="bottom")
    ranking["score"] = coexpression_unstack
    ranking["TF"] = list(ranking.index.get_level_values(level=0))
    ranking["TG"] = list(ranking.index.get_level_values(level=1))
    ranking = ranking.sort_values("score",ascending=False)
    ranking.index = ranking["TF"].astype(str)+"_"+ranking["TG"].astype(str)
    ranking = ranking.dropna()
    if pyscenic_format:
        ranking.rename(columns={'score':'importance',
                'TG':'target'},
                inplace=True)
        ranking = ranking.reset_index()
        ranking = ranking.drop(["rank","index"],axis=1)
        cols = ranking.columns.tolist()
        cols = ["TF","target","importance"]
        ranking = ranking[cols]
    return(ranking)

def join_rankings_scores_df(**rank_scores):
    """
    Join rankings and scores data frames generated by different methods.

    Args:
        **rank_scores: Named parameters, where arguments names should be the
            methods names and arguments values correspond to pandas.DataFrame
            output of rank_GRN

    Returns:
        (pandas.DataFrame, pandas.DataFrame): joined ranks and joined scores
            where rows represent possible regulatory links and columns represent
            each method.
            Values at row i and column j represent resp. the rank or the score
            of edge i computed by method j.

    Examples:
        >>> import pandas as pd
        >>> method1_rank = pd.DataFrame([[1,1.3, "gene1", "gene2"],
                                         [2,1.1, "gene1", "gene3"],
                                         [3,0.9, "gene3", "gene2"]],
                                         columns=['rank', 'score', 'TF', 'TG'])
        >>> method1_rank.index = method1_rank['TF']+'_'+method1_rank['TG']
        >>> method2_rank = pd.DataFrame([[1,1.4, "gene1", "gene3"],
                                         [2,1.0, "gene1", "gene2"],
                                         [3,0.9, "gene3", "gene2"]],
                                         columns=['rank', 'score', 'TF', 'TG'])
        >>> method2_rank.index = method2_rank['TF']+'_'+method2_rank['TG']
        >>> ranks, scores = join_rankings_scores_df(method1=method1_rank, method2=method2_rank)
        >>> ranks
                     method1  method2
        gene1_gene2        1        2
        gene1_gene3        2        1
        gene3_gene2        3        3
        >>> scores
                     method1  method2
        gene1_gene2      1.3      1.0
        gene1_gene3      1.1      1.4
        gene3_gene2      0.9      0.9

    """
    ranks_df = {method:rank_scores[method]["rank"] for method in rank_scores}
    ranks_df = pd.DataFrame(ranks_df)
    scores_df = {method:rank_scores[method]["score"] for method in rank_scores}
    scores_df = pd.DataFrame(scores_df)
    return(ranks_df,scores_df)



# -*- coding: utf-8 -*-
"""
This module allows to pre-process gene expression data.
"""
import numpy as np
import pandas as pd
try:
    import ot

except:
    print("Warning: could not import POT. columns_matrix_OT_norm() function cannot be called.")


def z_score(A, axis=0):
    """
    Compute the z-score along the specified axis.

    Args:
        A (pandas.DataFrame or numpy.array): matrix
        axis (int): 0 for columns and 1 for rows

    Returns:
        pandas.DataFrame or numpy.array: Normalized matrix

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> data = pd.DataFrame(np.random.randn(5, 5),
                            index=["c1", "c2", "c3", "c4", "c5"],
                            columns=["gene1", "gene2", "gene3", "gene4", "gene5"])
        >>> norm_data = z_score(data)
        >>> norm_data
               gene1     gene2     gene3     gene4     gene5
        c1  1.254757 -1.222682  0.914682  1.672581  0.828015
        c2 -0.446591 -0.083589 -1.038607 -0.418644 -0.331945
        c3  0.249333  0.960749  0.538403 -0.218012 -0.305461
        c4  0.367024  1.043200 -1.131598 -0.047267 -1.338834
        c5 -1.424523 -0.697678  0.717120 -0.988659  1.148225

    """
    epsilon = 1e-5
    if axis==0:
        A = (A - A.mean(axis=0)) / (A.std(axis=0) + epsilon)
    if axis==1:
        A = A.T
        A = (A - A.mean(axis=0)) / (A.std(axis=0) + epsilon)
        A = A.T
    return(A)

def mean_std_polishing(A, nb_iterations=5):
    """
    Iterative z-score on rows and columns.

    Args:
        A (pandas.DataFrame or numpy.array): matrix
        nb_iterations (int): number of polishing iterations

    Returns:
        pandas.DataFrame or numpy.array: Polished matrix

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> data = pd.DataFrame(np.random.randn(5, 5),
                            index=["c1", "c2", "c3", "c4", "c5"],
                            columns=["gene1", "gene2", "gene3", "gene4", "gene5"])
        >>> norm_data = mean_std_polishing(data)
        >>> norm_data
               gene1     gene2     gene3     gene4     gene5
        c1  0.336095 -1.618781  0.187436  1.109617 -0.014367
        c2 -0.321684  0.586608 -1.606905  0.484159  0.857821
        c3  0.139260  0.860934  0.976541 -1.395814 -0.580921
        c4  1.243263  0.421752 -0.585940  0.282319 -1.361394
        c5 -1.363323 -0.161066  0.826375 -0.421998  1.120013

    """
    epsilon = 1e-10
    for i in range(nb_iterations):
        A = A-A.mean(axis=0)
        # std polish the column
        A = A/(A.std(axis=0)+epsilon)
        # mean polish the row
        A = (A.T-A.T.mean(axis=0)).T
        # std polish the row
        A = (A.T/(A.T.std(axis=0) + epsilon)).T
    return(A)

def cat_gene_expression_dfs(gene_expression_dfs):
    """
    Concatenate different gene expression datasets, based on gene id (rows).

    Args:
        gene_expression_dfs (list of pandas.DataFrame): Expression datasets list

    Returns:
        pandas.DataFrame: concatenated gene expression datasets

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> data1 = pd.DataFrame(np.random.randn(3, 3),
                            index=["gene1", "gene2", "gene3"],
                            columns=["c1", "c2", "c3"])
        >>> data1
                     c1        c2        c3
        gene1  1.764052  0.400157  0.978738
        gene2  2.240893  1.867558 -0.977278
        gene3  0.950088 -0.151357 -0.103219
        >>> data2 = pd.DataFrame(np.random.randn(3, 3),
                            index=["gene2", "gene3", "gene4"],
                            columns=["c4", "c5", "c6"])
        >>> data2
                     c4        c5        c6
        gene2  0.410599  0.144044  1.454274
        gene3  0.761038  0.121675  0.443863
        gene4  0.333674  1.494079 -0.205158
        >>> data=cat_gene_expression_dfs([data1, data2])
        >>> data
                     c1        c2        c3        c4        c5        c6
        gene1  1.764052  0.400157  0.978738       NaN       NaN       NaN
        gene2  2.240893  1.867558 -0.977278  0.410599  0.144044  1.454274
        gene3  0.950088 -0.151357 -0.103219  0.761038  0.121675  0.443863
        gene4       NaN       NaN       NaN  0.333674  1.494079 -0.205158

    """
    cat = pd.concat(gene_expression_dfs,axis=1,sort=True)
    return(cat)

def median_outliers_filter(X, threshold=3):
    """
    Ensures that all the values of data_set are within:
    :math:`median(X) \pm \\tau \\times MAD(X))`

    Args:
        X (pandas.DataFrame or numpy.array): gene expression matrix (for instance)
        threshold (float): :math:`\\tau` threshold

    Returns:
        pandas.DataFrame or numpy.array: X without outliers (outliers set to
        the extreme values allowed)

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> data = pd.DataFrame(np.random.randn(5, 5),
                            index=["c1", "c2", "c3", "c4", "c5"],
                            columns=["gene1", "gene2", "gene3", "gene4", "gene5"])
        >>> median_outliers_filter(data)
               gene1     gene2     gene3     gene4     gene5
        c1  1.764052  0.400157  0.978738  0.674682  1.867558
        c2 -0.977278  0.950088 -0.653101 -0.103219  0.410599
        c3  0.144044  1.454274  0.761038  0.121675  0.443863
        c4  0.333674  1.494079 -0.653101  0.313068 -0.854096
        c5 -2.552990  0.653619  0.864436 -0.674682  2.269755

    """
    difference = np.abs(X-np.median(X,axis=0))
    median_difference = np.median(difference,axis=0)
    s = difference / median_difference
    mask = s > threshold
    mask_negative = X < np.median(X,axis=0)
    mask_positive = X > np.median(X,axis=0)
    X[np.logical_and(mask,mask_negative)] = -threshold*np.ones(X.shape)*median_difference
    X[np.logical_and(mask,mask_positive)] = threshold*np.ones(X.shape)*median_difference
    return(X)

def columns_matrix_OT_norm(X,reference=None,bins=None,**SinkhornTransport_para):
    """
    Use optimal transport in order to make all conditions disributions alike.

    Args:
        X (pandas.DataFrame): gene expression matrix
        r_percentile (numpy.array): reference distribution
        bins (numpy.array): bins for percentiles computation
        SinkhornTransport_para: ot.da.SinkhornTransport parameters

    Returns:
        pandas.DataFrame: Normalized matrix

    Examples:
        >>> import numpy as np
        >>> import pandas as pd
        >>> a = pd.DataFrame(np.random.randn(10000,10))
        >>> b = pd.DataFrame(np.random.randn(10000,10)*3+4)
        >>> bins = list(range(1,100))
        >>> b_ = columns_matrix_OT_norm(b,a.iloc[:,0],bins,reg_e=5e-1)
    """
    if bins is None:
        bins = [0.01]+list(np.arange(0.1,100,0.1))+[99.99]
    if reference is None:
        reference = X.values[:,0]
    percentile = np.percentile(reference,bins)
    X_c = pd.DataFrame()
    if "reg_e" not in SinkhornTransport_para:
        SinkhornTransport_para["reg_e"] = 1e-3
    for c in X.columns:
        percentile_c = np.percentile(X[c].values.flatten(),bins)
        ot_sinkhorn = ot.da.SinkhornTransport(**SinkhornTransport_para)
        ot_sinkhorn.fit(Xs=percentile_c.reshape(-1,1),
                        Xt=percentile.reshape(-1,1))
        X_c_transf = ot_sinkhorn.transform(Xs=X[c].values.reshape(-1,1))
        X_c[c] = X_c_transf.flatten()
    X_c.index = X.index
    return(X_c)



"""
GReNaDIne: Gene Regulatory Network Data-driven Inference

This package allows to infer Gene Regulatory Networks through several
Data-driven methods. Pre-processing and evaluation methods are also included.
"""
__author__ = "Sergio Peignier and Pauline Schmitt"
__copyright__ = "Copyright 2019, The GReNaDIne Project"
__license__ = "GPL"
__version__ = "0.1.8"
__maintainer__ = "Sergio Peignier"
__email__ = "sergio.peignier@insa-lyon.fr"
__status__ = "pre-alpha"



# -*- coding: utf-8 -*-
"""
This module allows to normalize RNAseq gene expression data.
"""
import pandas as pd
import numpy as np
try:
    from rpy2.robjects.packages import importr
    import rpy2.robjects as ro
    from rpy2.robjects import Formula
    from rpy2.robjects import pandas2ri
    deseq2 = importr('DESeq2')
    base = importr('base')
    biogeneric = importr('BiocGenerics')
    sum_exp = importr('SummarizedExperiment')
    pandas2ri.activate()

except:
    print("Warning: could not import R components. DEseq2() function cannot be called.")



def RPM(raw_counts):
    """
    Reads Per Million.

    Args:
        raw_counts (pandas.DataFrame): raw RNAseq counts where rows are genes
            and columns are conditions

    Returns:
        pandas.DataFrame: Normalized counts

    Examples:
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> import pandas as pd
        >>> nb_genes = 1000
        >>> nb_conditions = 5
        >>> raw_counts = np.random.randint(0,1e6,(nb_genes,nb_conditions))
        >>> raw_counts = pd.DataFrame(raw_counts)
        >>> rpm = RPM(raw_counts)
        >>> rpm.head()
                    0            1            2            3            4
        0  1994.031850   617.999738   895.038590   234.467249  1929.409246
        1   308.104674  1783.727269   738.867008   604.569366   245.491264
        2  1235.090833   906.128463   769.221953  1462.698984  1474.653899
        3   628.576824   344.838319  1723.115991  1201.585019  1084.225878
        4  1921.133736  1363.195297   761.440687  1481.020714  1777.679267

    """
    # per million factor: nb of reads per condition divided by 10^6
    pmf = raw_counts.sum(axis=0)/1e6
    rpm = raw_counts / pmf
    return(rpm)

def RPK(raw_counts, seq_lengths, seq_in_kb=False):
    """
    Reads Per Kilobase normalization.

    Args:
        raw_counts (pandas.DataFrame): raw RNAseq counts where rows are genes
            and columns are conditions
        seq_lengths (pandas.Series): sequences DNA lengths
        seq_in_kb (bool): True if lengths in kb, False otherwise

    Returns:
        pandas.DataFrame: Normalized counts

    Examples:
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> import pandas as pd
        >>> nb_genes = 1000
        >>> nb_conditions = 5
        >>> raw_counts = np.random.randint(0,1e6,(nb_genes,nb_conditions))
        >>> raw_counts = pd.DataFrame(raw_counts)
        >>> seq_lengths = np.random.randint(100,20000,nb_genes)
        >>> seq_lengths = pd.Series(seq_lengths)
        >>> rpk = RPK(raw_counts, seq_lengths)
        >>> rpk.head()
                       0              1              2              3              4
        0  321202.997719   99612.577387  142010.101010   38433.365917  313911.697621
        1   26853.843441  155566.114245   63431.417489   53620.768688   21611.248237
        2   97195.319962   71353.390640   59624.960204  117133.237822  117212.034384
        3  132006.796941   72465.590484  356436.703483  256785.896347  229981.733220
        4   48384.227419   34354.424576   18889.143614   37956.492944   45220.490091

    """
    seq_lengths = seq_lengths[raw_counts.index]
    seq_lengths = seq_lengths.fillna(seq_lengths.mean())
    if not seq_in_kb:
        seq_lengths_in_kb = seq_lengths / 1e3
    else:
        seq_lengths_in_kb = seq_lengths
    rpk = (raw_counts.T / seq_lengths_in_kb).T
    return(rpk)

def RPKM(raw_counts, seq_lengths, seq_in_kb=False):
    """
    Reads Per Kilobase Million (also known as FPM: Fragments per kilobase).

    Args:
        raw_counts (pandas.DataFrame): raw RNAseq counts where rows are genes
            and columns are conditions
        seq_lengths (pandas.Series): sequences DNA lengths
        seq_in_kb (bool): True if lengths in kb, False otherwise

    Returns:
        pandas.DataFrame: Normalized counts

    Examples:
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> import pandas as pd
        >>> nb_genes = 1000
        >>> nb_conditions = 5
        >>> raw_counts = np.random.randint(0,1e6,(nb_genes,nb_conditions))
        >>> raw_counts = pd.DataFrame(raw_counts)
        >>> seq_lengths = np.random.randint(100,20000,nb_genes)
        >>> seq_lengths = pd.Series(seq_lengths)
        >>> rpkm = RPKM(raw_counts, seq_lengths)
        >>> rpkm.head()
                    0           1           2           3           4
        0  649.733415  201.368439  291.638511   76.398582  628.676848
        1   54.320288  314.479420  130.265692  106.588393   43.281252
        2  196.607901  144.242035  122.448576  232.839698  234.742741
        3  267.024989  146.490365  731.994898  510.443933  460.588733
        4   97.872216   69.448026   38.791619   75.450645   90.563924

    """
    # reads per million
    rpm = RPM(raw_counts=raw_counts)
    rpkm = RPK(rpm, seq_lengths, seq_in_kb=False)
    return(rpkm)

def TPM(raw_counts, seq_lengths, seq_in_kb=False):
    """
    Transcript Per Million normalization.

    Args:
        raw_counts (pandas.DataFrame): raw RNAseq counts where rows are genes
            and columns are conditions
        seq_lengths (pandas.Series): sequences DNA lengths
        seq_in_kb (bool): True if lengths in kb, False otherwise

    Returns:
        pandas.DataFrame: Normalized counts

    Examples:
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> import pandas as pd
        >>> nb_genes = 1000
        >>> nb_conditions = 5
        >>> raw_counts = np.random.randint(0,1e6,(nb_genes,nb_conditions))
        >>> raw_counts = pd.DataFrame(raw_counts)
        >>> seq_lengths = np.random.randint(100,20000,nb_genes)
        >>> seq_lengths = pd.Series(seq_lengths)
        >>> tpm = TPM(raw_counts, seq_lengths)
        >>> tpm.head()
                     0            1            2            3            4
        0  2455.468465   739.530213  1103.147117   265.510632  2397.256398
        1   205.286894  1154.932887   492.740902   370.430324   165.039097
        2   743.019352   529.732184   463.172003   809.195846   895.115733
        3  1009.139172   537.989227  2768.832068  1773.963432  1756.306584
        4   369.878069   255.049468   146.732550   262.216233   345.336316

    """
    rpk = RPK(raw_counts, seq_lengths, seq_in_kb)
    tpm = RPM(rpk)
    return(tpm)


def DEseq2(raw_counts,col_data,rlog=True):
    """
    Apply R DEseq2 normalization.

    Args:
        raw_counts (pandas.DataFrame): raw RNAseq counts where rows are genes
            and columns are conditions
        col_data (pandas.DataFrame): Two columns, one corresponding to ids of
            each condition (individuals), and one with the experiment id
            (if many repetitions)

    Returns:
        pandas.DataFrame: Normalized counts

    Example:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> raw_counts = pd.DataFrame(np.random.randint(0,1000,(20,10)),
                                      columns = ["Z"+str(i) for i in range(10)])
        >>> col_data = pd.DataFrame([["Z0","1"],
                                     ["Z1","2"],
                                     ["Z2","3"],
                                     ["Z3","4"],
                                     ["Z4","5"],
                                     ["Z5","6"],
                                     ["Z6","7"],
                                     ["Z7","8"],
                                     ["Z8","9"],
                                     ["Z9","10"]
                                     ],columns=["individuals","conditions"])
        >>> raw_counts.columns = col_data["individuals"]
        >>> col_data.index = col_data['individuals']
        >>> DEseq2(raw_counts,col_data,rlog=False)
        individuals          X0          X1     ...              X8          X9
        0            408.025477  382.991634     ...        7.745300  611.474516
        1            165.238388  516.593367     ...      270.224902  596.251084
        2            289.912839  377.510537     ...      727.197585   60.893728
        3            463.502625  627.585575     ...      385.543809  718.884285
        4             59.056319  674.174898     ...      364.029087  243.574911
        5            573.263865  181.561329     ...      129.948918  570.878697
        6            304.229522  314.477925     ...      802.068816   44.824550
        7            537.472156  376.825400     ...       36.144732  373.819828
        8            323.914962  608.401737     ...      748.712307  100.643800
        9            464.695682  294.608949     ...      781.414683  535.357356
        10           559.543710   57.551516     ...      112.737140  822.065324
        11           517.786716  123.324676     ...      618.763389  768.783312
        12           222.505121  584.421939     ...       81.755942  166.612005
        13           361.496256  175.395095     ...      333.047888  515.905193
        14           330.476775  666.638390     ...      779.693506  312.926101
        15           331.073304  653.620785     ...      493.978005  787.389729
        16           437.851901   84.271862     ...      483.650938  347.601696
        17           466.485268   28.090621     ...      750.433484    9.303208
        18           459.326926  210.337087     ...      149.742461  468.543405
        19           221.312064  126.065225     ...      662.653421  435.559302

    """
    if len(col_data.index) != len(raw_counts.columns):
        print('each raw of col_data should correspond to each column of raw_counts ... dataframes shapes mismatches')
        return(1)
    if (col_data.index!=raw_counts.columns).any():
        col_data.index = raw_counts.columns
    dds = deseq2.DESeqDataSetFromMatrix(raw_counts,col_data,Formula('~ conditions'))
    dds = deseq2.DESeq(dds)
    res = deseq2.results(dds)
    if rlog:
        rlog_counts = deseq2.rlog(dds)
        normed_counts = sum_exp.assay(rlog_counts)
    else:
        normed_counts = biogeneric.counts(dds,True)
    normed_counts = pd.DataFrame(normed_counts)
    normed_counts.index = raw_counts.index
    normed_counts.columns = raw_counts.columns
    return(normed_counts)

def log(X,base=10, pseudocount=1):
    """
    Add a pseudocount and apply the log transformation with a given base.

    Args:
        X (pandas.DataFrame or numpy.array): gene expression matrix
        base (float): logarithm base
        pseudocount(float): pseudocount value

    Returns:
        pandas.DataFrame or numpy.array: log transformed gene expression matrix

    Examples:
        >>> import pandas as pd
        >>> import numpy as np
        >>> np.random.seed(0)
        >>> data = pd.DataFrame(np.random.randn(5, 5),
                            index=["c1", "c2", "c3", "c4", "c5"],
                            columns=["gene1", "gene2", "gene3", "gene4", "gene5"])
        >>> pseudocount = -np.min(data.values)+1
        >>> log_data = log(data, pseudocount=pseudocount)
        >>> log_data
               gene1     gene2     gene3     gene4     gene5
        c1  0.725670  0.596943  0.656264  0.762970  0.734043
        c2  0.410897  0.653509  0.531687  0.537790  0.598089
        c3  0.567853  0.699600  0.634883  0.565218  0.601718
        c4  0.589577  0.703039  0.524764  0.587268  0.431186
        c5  0.000000  0.623932  0.645169  0.448834  0.765128

    """
    X_log = np.log(X+pseudocount)
    return(X_log/np.log(base))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="GReNaDIne",
    version="0.1.8",
    author="Sergio Peignier, Pauline Schmitt, Timothée Frouté",
    author_email="sergio.peignier@insa-lyon.fr",
    description="Data-Driven Gene Regulatory Network Inference",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    install_requires=[
        'pandas',
        'numpy',
        'scikit-learn',
        'tqdm',
        'scipy',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
)

print(setuptools.find_packages())
